/* NAT for netfilter; shared with compatibility layer. */

/* (c) 1999 Paul `Rusty' Russell.  Licenced under the GNU General
   Public Licence. */
#include <linux/types.h>
#include <linux/timer.h>
#include <linux/skbuff.h>
#include <linux/netfilter_ipv4.h>
#include <net/checksum.h>
#include <net/icmp.h>
#include <net/ip.h>
#include "ip_nat.h"
#include "ip_nat_protocol.h"
#include "ip_nat_core.h"

size_t ip_nat_offset;

struct ip_nat_protocol *protos;
extern struct ip_nat_protocol unknown_nat_protocol;

struct ip_nat_hash *bysource[IP_NAT_HTABLE_SIZE] = { NULL };
struct ip_nat_hash *byipsproto[IP_NAT_HTABLE_SIZE] = { NULL };

static void ip_nat_cleanup_state(struct ip_state *state)
{
	struct ip_nat_info *info = (void *)state + ip_nat_offset;

	DEBUGP("ip_nat_cleanup_state: cleaning up %p\n", state);
	if (info->bysource.state) {
		unsigned int hash
			= hash_by_src(&state->tuplehash[0].tuple.src,
				      state->tuplehash[0].tuple.dst.protonum);
		modreg_delete(&bysource[hash], &info->bysource);
	}

	if (info->byipsproto.state) {
		unsigned int hash
			= hash_by_ipsproto(state->tuplehash[1].tuple.src.ip,
					   state->tuplehash[1].tuple.dst.ip,
					   state->tuplehash[1].tuple.dst
					   .protonum);
		modreg_delete(&byipsproto[hash], &info->byipsproto);
	}
}

/* We do checksum mangling, so if they were wrong before they're still
 * wrong.  Also works for incomplete packets (eg. ICMP dest
 * unreachables.) */
u_int16_t 
ip_nat_cheat_check(u_int32_t oldvalinv, u_int32_t newval, u_int16_t oldcheck)
{
	u_int32_t diffs[] = { oldvalinv, newval };
	return csum_fold(csum_partial((char *)diffs, sizeof(diffs),
				      oldcheck^0xFFFF));
}

struct ip_nat_protocol *
find_nat_proto(u_int16_t protonum)
{
	struct ip_nat_protocol *i;

	for (i = protos; i; i = i->next) 
		if (i->protonum == protonum)
			return i;

	return &unknown_nat_protocol;
}

/* Is this tuple already taken? (not by us)*/
int
ip_nat_used_tuple(const struct ip_state_tuple *tuple,
		  const struct ip_state *ignored_state)
{
	/* State tracking doesn't keep track of outgoing tuples; only
	   incoming ones.  NAT means they don't have a fixed mapping,
	   so we invert the tuple and look for the incoming reply.

	   We could keep a separate hash if this proves too slow. */
	struct ip_state_tuple reply;

	invert_tuplepr(&reply, tuple);
	return (int)ip_state_find(&reply, ignored_state);
}

/* Does tuple + the source manip come within the range mr */
static int
in_range(const struct ip_state_tuple *tuple,
	 const struct ip_state_manip *manip,
	 const struct ip_nat_multi_range *mr)
{
	struct ip_nat_protocol *proto = find_nat_proto(tuple->dst.protonum);
	unsigned int i;
	struct ip_state_tuple newtuple = { *manip, tuple->dst };

	for (i = 0; i < mr->rangesize; i++) {
		/* If we are allowed to map IPs, then we must be in the
		   range specified, otherwise we must be unchanged. */
		if (mr->range[i].flags & IP_NAT_RANGE_MAP_IPS) {
			if (ntohl(newtuple.src.ip) < ntohl(mr->range[i].min_ip)
			    || (ntohl(newtuple.src.ip)
				> ntohl(mr->range[i].max_ip)))
				continue;
		} else {
			if (newtuple.src.ip != tuple->src.ip)
				continue;
		}

		if ((mr->range[i].flags & IP_NAT_RANGE_PROTO_SPECIFIED)
		    && proto->in_range(&newtuple, IP_NAT_MANIP_SRC,
				       &mr->range[i].min, &mr->range[i].max))
			return 1;
	}
	return 0;
}

/* Only called for SRC manip */
static struct ip_state_manip *
find_appropriate_src(const struct ip_state_tuple *tuple,
		     const struct ip_nat_multi_range *mr)
{
	unsigned int hash = hash_by_src(&tuple->src, tuple->dst.protonum);
	struct ip_nat_hash *i;

	for (i = bysource[hash]; i; i = i->next) {
		if (i->state->tuplehash[0].tuple.dst.protonum
		    == tuple->dst.protonum
		    && i->state->tuplehash[0].tuple.src.ip
		    == tuple->src.ip
		    && i->state->tuplehash[0].tuple.src.u.all
		    == tuple->src.u.all
		    && in_range(tuple, &i->state->tuplehash[0].tuple.src, mr))
			return &i->state->tuplehash[0].tuple.src;
	}
	return 0;
}

/* If it's really a local destination manip, it may need to do a
   source manip too. */
static int
do_extra_mangle(u_int32_t var_ip, u_int32_t *other_ipp)
{
	struct rtable *rt;

	/* FIXME: IPTOS_TOS(iph->tos) --RR */
	if (ip_route_output(&rt, var_ip, 0, 0, 0) != 0) {
		DEBUGP("do_extra_mangle: Can't get route to %u.%u.%u.%u\n",
		       IP_PARTS(var_ip));
		return 0;
	}

	DEBUGP("do_extra_mangle: Replacing src %u.%u.%u.%u with %u.%u.%u.%u\n",
	       IP_PARTS(*other_ipp), IP_PARTS(rt->rt_src));
	*other_ipp = rt->rt_src;
	ip_rt_put(rt);
	return 1;
}

static inline unsigned int
count_maps(u_int32_t src, u_int32_t dst, u_int16_t protonum)
{
	struct ip_nat_hash *h;
	unsigned int score = 0;

	for (h = byipsproto[hash_by_ipsproto(src, dst, protonum)];
	     h; h = h->next) {
		if (h->state->tuplehash[1].tuple.src.ip == src
		    && h->state->tuplehash[1].tuple.dst.ip == dst
		    && h->state->tuplehash[1].tuple.dst.protonum == protonum)
			score++;
	}

	return score;
}

/* For [FUTURE] fragmentation handling, we want the least-used
   src-ip/dst-ip/proto triple.  Fairness doesn't come into it.  Thus
   if the range specifies 1.2.3.4 ports 10000-10005 and 1.2.3.5 ports
   1-65535, we don't do pro-rata allocation based on ports; we choose
   the ip with the lowest src-ip/dst-ip/proto usage.

   If an allocation then fails (eg. all 6 ports used in the 1.2.3.4
   range), we eliminate that and try again.  This is not the most
   efficient approach, but if you're worried about that, don't hand us
   ranges you don't really have.  */
static struct ip_nat_range *
find_best_ips_proto(struct ip_state_tuple *tuple,
		    const struct ip_nat_multi_range *mr,
		    enum ip_nat_manip_type maniptype,
		    const struct device *device)
{
	unsigned int i;
	struct {
		const struct ip_nat_range *range;
		unsigned int score;
		struct ip_state_tuple tuple;
	} best = { NULL,  0xFFFFFFFF };
	u_int32_t *var_ipp, *other_ipp, saved_ip;

	if (maniptype == IP_NAT_MANIP_SRC) {
		DEBUGP("find_best_ips_proto: SRC manip.\n");
		var_ipp = &tuple->src.ip;
		saved_ip = tuple->dst.ip;
		other_ipp = &tuple->dst.ip;
	} else {
		DEBUGP("find_best_ips_proto: DST manip.\n");
		var_ipp = &tuple->dst.ip;
		saved_ip = tuple->src.ip;
		other_ipp = &tuple->src.ip;
	}

	for (i = 0; i < mr->rangesize; i++) {
		u_int32_t minip, maxip;

		/* Don't do ranges which are already eliminated. */
		if (mr->range[i].flags & IP_NAT_RANGE_FULL) continue;

		if (mr->range[i].flags & IP_NAT_RANGE_MAP_IPS) {
			minip = mr->range[i].min_ip;
			maxip = mr->range[i].max_ip;
		} else
			minip = maxip = *var_ipp;

		DEBUGP("ip_nat_find_best: %u.%u.%u.%u - %u.%u.%u.%u\n",
		       IP_PARTS(minip), IP_PARTS(maxip));
		for (*var_ipp = minip;
		     ntohl(*var_ipp) <= ntohl(maxip);
		     *var_ipp = htonl(ntohl(*var_ipp) + 1)) {
			unsigned int score;

			/* Reset the other ip in case it was mangled by
			 * do_extra_mangle last time. */
			*other_ipp = saved_ip;

			if (!device
			    && maniptype == IP_NAT_MANIP_DST
			    && !do_extra_mangle(*var_ipp, other_ipp)) {
				/* Can't route?  This whole range part is
				 * probably screwed, but keep trying
				 * anyway. */
				continue;
			}

			/* Count how many others map onto this. */
			score = count_maps(tuple->src.ip, tuple->dst.ip,
					   tuple->dst.protonum);

			DEBUGP("ip_nat_find_best: score for %u.%u.%u.%u: %u\n",
			       IP_PARTS(*var_ipp), score);
			if (score < best.score) {
				DEBUGP("ip_nat_find_best_ips_proto: score %u "
				       "beats %u ip=%u.%u.%u.%u\n",
				       score, best.score,
				       IP_PARTS(*var_ipp));
				/* Optimization: doesn't get any better than
				   this. */
				if (score == 0)
					return (struct ip_nat_range *)
						&mr->range[i];

				best.score = score;
				best.tuple = *tuple;
				best.range = &mr->range[i];
			}
		}
	}
	*tuple = best.tuple;

	/* Discard const. */
	return (struct ip_nat_range *)best.range;
}

static int
get_unique_tuple(struct ip_state_tuple *tuple,
		 const struct ip_state_tuple *orig_tuple,
		 enum ip_nat_manip_type maniptype,
		 struct ip_nat_multi_range *mr,
		 const struct ip_state *state,
		 const struct device *device)
{
	struct ip_nat_protocol *proto
		= find_nat_proto(orig_tuple->dst.protonum);
	struct ip_nat_range *rptr;
	unsigned int i;

	/* 1) If this srcip/proto/src-proto-part is currently mapped,
	   and that same mapping gives a unique tuple within the given
	   range, use that.

	   This is only required for source (ie. NAT/masq) mappings.
	   So far, we don't do local source mappings, so multiple
	   manips not an issue.  */
	if (maniptype == IP_NAT_MANIP_SRC) {
		struct ip_state_manip *manip;

		manip = find_appropriate_src(orig_tuple, mr);
		if (manip) {
			/* Apply same source manipulation. */
			*tuple = ((struct ip_state_tuple)
				  { *manip, orig_tuple->dst });
			
			DEBUGP("get_unique_tuple: Found current mapping\n");
			return 1;
		}
	}

	/* Clear full flags. */
	for (i = 0; i < mr->rangesize; i++)
		mr->range[i].flags &= ~IP_NAT_RANGE_FULL;

	/* 2) Select the least-used IP/proto combination in the given
	   range.
	*/
	*tuple = *orig_tuple;
	while ((rptr = find_best_ips_proto(tuple, mr, maniptype, device))
	       != NULL) {
		/* 3) The per-protocol part of the manip is made to
		   map into the range to make a unique tuple. */

		/* Only bother mapping if it's not already in range
		   and unique */
		if ((!(rptr->flags & IP_NAT_RANGE_PROTO_SPECIFIED)
		     || proto->in_range(tuple, maniptype,
					&rptr->min, &rptr->max))
		    && !ip_nat_used_tuple(tuple, state)) {
			DEBUGP("No need to remap per-proto.\n");
			return 1;
		}
		else if (proto->unique_tuple(tuple, rptr, maniptype, state)) {
#ifdef DEBUGNAT
			if (ip_nat_used_tuple(tuple, state))
				DEBUGP("Proto made nonunique tuple!\n");
#endif
			return 1;
		}

		/* Eliminate that from range, and try again. */
		rptr->flags |= IP_NAT_RANGE_FULL;
		*tuple = *orig_tuple;
	}

	return 0;
}

unsigned int
ip_nat_setup_info(struct ip_state *state,
		  enum ip_nat_manip_type maniptype,
		  struct ip_nat_multi_range *mr,
		  const struct device *device)
{
	struct ip_state_tuple new_tuple, inv_tuple, reply;
	struct ip_nat_manipulation *m, *m_rev;
	struct ip_nat_info *info = (void *)state + ip_nat_offset;

	static unsigned int opposite[]
		= { [NF_IP_LOCAL_OUT] = NF_IP_LOCAL_IN,
		    [NF_IP_LOCAL_IN] = NF_IP_LOCAL_OUT,
		    [NF_IP_POST_ROUTING] = NF_IP_PRE_ROUTING,
		    [NF_IP_PRE_ROUTING] = NF_IP_POST_ROUTING
		};

	/* Set up which hooks manips are to be done on. */
	/* Local packet, dst manip: NF_IP_LOCAL_OUT.
	   Local packet, src manip: NF_IP_POST_ROUTING.
	   Non-Local packet: dst manip: NF_IP_PRE_ROUTING.
	   Non-Local packet: src manip: NF_IP_POST_ROUTING.
	*/
	if (maniptype == IP_NAT_MANIP_SRC) 
		info->dirinfo[0].hook = NF_IP_POST_ROUTING;
	else if (!device) 
		info->dirinfo[0].hook = NF_IP_LOCAL_OUT;
	else
		info->dirinfo[0].hook = NF_IP_PRE_ROUTING;

	info->dirinfo[1].hook = opposite[info->dirinfo[0].hook];

	m = info->dirinfo[0].manipulation;
	m_rev = info->dirinfo[1].manipulation;

	if (!get_unique_tuple(&new_tuple, &state->tuplehash[0].tuple,
			      maniptype, mr, state, device)) {
		DEBUGP("ip_nat_alloc_binding: Can't get unique tuple.\n");
		return NF_DROP;
	}

	/* We now have two tuples (SRCIP/SRCPT/DSTIP/DSTPT):
	   the original (A/B/C/D') and the mangled one (E/F/G/H').

	   We're only allowed to work with the SRC per-proto part,
	   so we create inverses of both to start, then derive the other
	   fields we need.
	*/

	/* Reply connection: simply invert the new tuple (G/H/E/F') */
	invert_tuplepr(&reply, &new_tuple);

	/* Alter state table so it recognizes replies. */
	ip_state_alter_reply(state, &reply);

	/* Create inverse of original: C/D/A/B' */
	invert_tuplepr(&inv_tuple, &state->tuplehash[0].tuple);

	/* E != A or F != B => a source manip. */
	if (memcmp(&new_tuple.src, &state->tuplehash[0].tuple.src,
		   sizeof(new_tuple.src)) != 0) {
		DEBUGP("new_tuple.src != orig.src: %u.%u.%u.%u:%u v %u.%u.%u.%u:%u\n",
		       IP_PARTS(new_tuple.src.ip), new_tuple.src.u.icmp.id,
		       IP_PARTS(state->tuplehash[0].tuple.src.ip),
		       state->tuplehash[0].tuple.src.u.icmp.id);
		DEBUGP("Manips so far: %u and %u\n",
		       info->dirinfo[0].num_manips,
		       info->dirinfo[1].num_manips);

		/* In this direction, a source manip. */
		m->maniptype = IP_NAT_MANIP_SRC;
		m->manip = new_tuple.src;
		m++;
		info->dirinfo[0].num_manips++;

		/* In the reverse direction, a destination manip. */
		m_rev->maniptype = IP_NAT_MANIP_DST;
		m_rev->manip = state->tuplehash[0].tuple.src;
		m_rev++;
		info->dirinfo[1].num_manips++;
	}

	/* G != C or H != D => a destination manip. */
	if (memcmp(&reply.src, &inv_tuple.src, sizeof(inv_tuple.src)) != 0) {
		DEBUGP("reply.src != inv_tuple.src: %u.%u.%u.%u:%u v %u.%u.%u.%u:%u\n",
		       IP_PARTS(reply.src.ip), reply.src.u.icmp.id,
		       IP_PARTS(inv_tuple.src.ip), inv_tuple.src.u.icmp.id);

		/* In this direction, a destination manip */
		m->maniptype = IP_NAT_MANIP_DST;
		m->manip = reply.src;
		m++;
		info->dirinfo[0].num_manips++;

		/* In the reverse direction, a source manip. */
		m_rev->maniptype = IP_NAT_MANIP_SRC;
		m_rev->manip = inv_tuple.src;
		m_rev++;
		info->dirinfo[1].num_manips++;
	}
	return NF_ACCEPT;
}

static void
manip_pkt(u_int16_t proto, struct iphdr *iph, size_t len,
	  const struct ip_state_manip *manip,
	  enum ip_nat_manip_type maniptype)
{
	find_nat_proto(proto)->manip_pkt(iph, len, manip, maniptype);

	if (maniptype == IP_NAT_MANIP_SRC) {
		DEBUGP("IP: Replacing SRC %u.%u.%u.%u with %u.%u.%u.%u\n",
		       IP_PARTS(iph->saddr), IP_PARTS(manip->ip));
		iph->check = ip_nat_cheat_check(~iph->saddr, manip->ip,
						iph->check);
		iph->saddr = manip->ip;
	} else {
		DEBUGP("IP: Replacing DST %u.%u.%u.%u with %u.%u.%u.%u\n",
		       IP_PARTS(iph->daddr), IP_PARTS(manip->ip));
		iph->check = ip_nat_cheat_check(~iph->daddr, manip->ip,
						iph->check);
		iph->daddr = manip->ip;
	}
#if 0
	if (ip_fast_csum((u8 *)iph, iph->ihl) != 0)
		DEBUGP("IP: checksum on packet bad.\n");

	if (proto == IPPROTO_TCP) {
		void *th = (u_int32_t *)iph + iph->ihl;
		if (tcp_v4_check(th, len - 4*iph->ihl, iph->saddr, iph->daddr,
				 csum_partial((char *)th, len-4*iph->ihl, 0)))
			DEBUGP("TCP: checksum on packet bad\n");
	}
#endif
}

/* Do packet manipulations according to binding. */
void
do_binding(struct ip_nat_info *info, int dir, struct iphdr *iph, size_t len)
{
	unsigned int i;

	for (i = 0; i < info->dirinfo[dir].num_manips; i++) {
#if 1
		DEBUGP("Manipulation %u: dir %s type %s (ip = %u.%u.%u.%u)\n",
		       i, dir ? "reply" : "orig",
		       info->dirinfo[dir].manipulation[i].maniptype
		       == IP_NAT_MANIP_SRC ? "SRC" : "DST",
		       IP_PARTS(info->dirinfo[dir].manipulation[i].manip.ip));
#endif
		manip_pkt(iph->protocol, iph, len,
			  &info->dirinfo[dir].manipulation[i].manip,
			  info->dirinfo[dir].manipulation[i].maniptype);
	}
}

void 
icmp_reply_translation(struct sk_buff *skb, 
		       struct ip_state *state,
		       int conn_dir)
{
	struct iphdr *iph = skb->nh.iph;
	struct icmphdr *hdr = (struct icmphdr *)((u_int32_t *)iph + iph->ihl);
	struct iphdr *inner = (struct iphdr *)(hdr + 1);
	struct ip_nat_info *info = (void *)state + ip_nat_offset;
	size_t datalen = skb->len - ((void *)inner - (void *)iph);
	unsigned int i;

	/* Map inner packet first; do this REVERSE of normal. */
	for (i = 0; i < info->dirinfo[conn_dir].num_manips; i++) {
		manip_pkt(inner->protocol, inner, 
			  skb->len - ((void *)inner - (void *)iph),
			  &info->dirinfo[conn_dir].manipulation[i].manip,
			  !info->dirinfo[conn_dir].manipulation[i].maniptype);
	}

	/* Now outer packet. */
	/* Note: May not be from a NAT'd host, but probably safest to
	   do translation always as if it came from the host itself
	   (even though a "host unreachable" coming from the host
	   itself is a bit wierd).

	   More explanation: some people use NAT for anonomising.
	   Also, CERT recommends dropping all packets from private IP
	   addresses (although ICMP errors from internal links with
	   such addresses are not too uncommon, as Alan Cox points
	   out) */

	/* Use mapping to map outer packet: 0 give no per-proto mapping */
	for (i = 0; i < info->dirinfo[conn_dir].num_manips; i++) {
		manip_pkt(0, iph, skb->len,
			  &info->dirinfo[conn_dir].manipulation[i].manip,
			  info->dirinfo[conn_dir].manipulation[i].maniptype);
	}

	/* Since we mangled inside ICMP packet, recalculate its
	   checksum from scratch.  (Hence the handling of incorrect
	   checksums above, so we don't accidentally fix one.) */
	hdr->checksum = 0;
	hdr->checksum = ip_compute_csum((unsigned char *)hdr,
					sizeof(*hdr) + datalen);
}

void ip_nat_init(void)
{
	/* Reserve space in state structure. */
	ip_nat_offset = struct_reserve(ip_state, 
				       sizeof(struct ip_nat_info));

	/* Sew in builtin protocols. */
	protos = &ip_nat_protocol_tcp;
	ip_nat_protocol_tcp.next = &ip_nat_protocol_udp;
	ip_nat_protocol_udp.next = &ip_nat_protocol_icmp;

	IP_STATE_ASSERT(ip_state_destroyed == NULL);
	ip_state_destroyed = &ip_nat_cleanup_state;
}
	
void ip_nat_cleanup(void)
{
	/* Unreserve space in state structure. */
	struct_unreserve(ip_state, ip_nat_offset, sizeof(struct ip_nat_info));

	ip_state_destroyed = NULL;
}
