/* Everything about the rules for NAT. */
#define __NO_VERSION__
#include <linux/config.h>
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/module.h>
#include <linux/kmod.h>
#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <net/checksum.h>
#include <asm/spinlock.h>
#include <linux/bitops.h>

#include "ip_nat_rule.h"
#include "ip_nat_protocol.h"

static struct ip_nat_rule *nat_rules = NULL;
static struct ip_nat_mapping_type *mapping_types = NULL;

struct ip_nat_rule *
find_rule(const struct ip_state_tuple *tuple)
{
	struct ip_nat_rule *ptr;

	for (ptr = nat_rules; ptr; ptr = ptr->next) {
		unsigned int i;
		unsigned long cmp = 0;

		/* Apply mask.  Should unroll nicely */
		for (i = 0; i < sizeof(*tuple)/sizeof(unsigned long); i++) {
			cmp |= (((unsigned long *)tuple)[i]
				& ((unsigned long *)&ptr->mask)[i])
				^ ((unsigned long *)&ptr->match)[i];
		}

		if (cmp == 0) break;
	}

	return ptr;
}

/* NAT rule tables are ordered, similar to routing tables: most
 * specific to least specific.  I've tried (and maybe failed) to be
 * intuitive here.  Higher the better.  This is a partial ordering.  */
static u_int32_t
precedence(const struct ip_state_tuple *mask)
{
	u_int32_t prec = 0;

	/* We build this up most important to least. */

	/* Source IP bits + destination IP bits: 7 bits */
	prec |= hweight32(mask->src.ip) + hweight32(mask->dst.ip);

	/* Do we care about protocol?: 1 bit */
	prec <<= 1;
	prec |= mask->dst.protonum ? 1 : 0;

	/* Per-protocol information: 4 bits */
	prec <<= 4;
	prec |= hweight16(mask->src.u.all) + hweight16(mask->dst.u.all);

	return prec;
}

static struct ip_nat_mapping_type *
find_mapping_type(const char *name, int load)
{
	struct ip_nat_mapping_type *ret;

	ret = modreg_named_find(&mapping_types, name);
	if (!ret && load) {
		char modulename[IP_NAT_MAPPING_TYPE_MAX_NAMELEN
			       + sizeof("ip_nat_map_")];
		strcpy(modulename, "ip_nat_map_");
		strcat(modulename, name);
		DEBUGP("ip_nat: requesting module `%s'\n", modulename);
		request_module(modulename);
		return find_mapping_type(name, 0);
	}
	else return ret;
}

int ip_nat_register_mapping_type(struct ip_nat_mapping_type *me)
{
	if (!modreg_named_insert(&mapping_types, me))
		return -EEXIST;
	MOD_INC_USE_COUNT;
	return 0;
}

void ip_nat_unregister_mapping_type(struct ip_nat_mapping_type *me)
{
	modreg_delete(&mapping_types, me);
	MOD_DEC_USE_COUNT;
}

/* Return 0 or -errno */
static int rule_user_to_rule(const struct ip_nat_rule_user *user,
			     struct ip_nat_rule **rule,
			     struct ip_nat_mapping_type *mtype)
{
	/* FIXME: check range and per-proto parts of tuples only set
           if a specific protocol mentioned. --RR */
	if (!user->mask.dst.protonum
	    && (user->mask.src.u.all || user->mask.dst.u.all
		|| (user->range.flags & IP_NAT_RANGE_PROTO_SPECIFIED))) {
		DEBUGP("rule_user_to_rule: bad proto specifics\n");
		return -EINVAL;
	}

	*rule = kmalloc(sizeof(struct ip_nat_rule), GFP_USER);
	if (!*rule)
		return -ENOMEM;

	**rule = ((struct ip_nat_rule){ NULL, user->match, user->mask,
					user->maniptype,
					precedence(&user->mask), mtype,
					{ 1, { user->range } } });

	/* Make sure padding is zero. */
	(*rule)->match.src.pad = (*rule)->mask.src.pad = 0;

	if (mtype && !mtype->check(*rule)) {
		kfree(*rule);
		return -EINVAL;
	}
	return 0;
}

static int prec_cmpfn(const void *new, const void *curr)
{
	return ((struct ip_nat_rule *)new)->precedence
		> ((struct ip_nat_rule *)curr)->precedence;
}

/* FIXME: insert should add to range if rule already exists, and
   delete should delete from range. --RR */
static int insert_rule(struct ip_nat_rule_user *ruleuser,
		       struct ip_nat_mapping_type *mtype)
{
	int ret;
	struct ip_nat_rule *rule;

	ret = rule_user_to_rule(ruleuser, &rule, mtype);
	if (ret == 0) {
		modreg_insert(&nat_rules, rule, prec_cmpfn);
		if (mtype && mtype->me)
			__MOD_INC_USE_COUNT(mtype->me);
	}
	return ret;
}

static int
rule_cmpfn(const void *find, const void *curr)
{
	/* Compare everything after next ptr in elem. */
	return memcmp(find + sizeof(struct ip_nat_rule *), 
		      curr + sizeof(struct ip_nat_rule *),
		      sizeof(struct ip_nat_rule) 
		      - sizeof(struct ip_nat_rule *))
		== 0;
}

static int delete_rule(const struct ip_nat_rule_user *ruleuser,
		       struct ip_nat_mapping_type *mtype)
{
	int ret;
	struct ip_nat_rule *rule;

	ret = rule_user_to_rule(ruleuser, &rule, mtype);
	if (ret == 0) {
		struct ip_nat_rule *match;
		match = modreg_find(&nat_rules, rule, rule_cmpfn);
		if (!match)
			ret = -ENOENT;
		else {
			modreg_delete(&nat_rules, match);
			if (mtype && mtype->me)
				__MOD_DEC_USE_COUNT(mtype->me);
			kfree(rule);
			ret = 0;
		}
		
	}
	return ret;
}

static int do_ip_nat_ctl(int cmd, void *user, unsigned int len)
{
	int ret;
	struct ip_nat_rule_user ruleuser;

	if (len != sizeof(struct ip_nat_rule_user)) {
		DEBUGP("IP_NAT_SO_INSERT/DELETE: len %u != %u\n",
		       len, sizeof(struct ip_nat_rule_user));
		ret = -EINVAL;
	} else if (copy_from_user(&ruleuser, user, len)) {
		ret = -EFAULT;
	} else {
		struct ip_nat_mapping_type *mtype = NULL;

		/* Must be NUL terminated for sanity */
		ruleuser.typename[IP_NAT_MAPPING_TYPE_MAX_NAMELEN-1] = '\0';

		/* Need to do this outside lock. */
		if (!ruleuser.typename[0]
		    || (mtype = find_mapping_type(ruleuser.typename, 1))) {
			unsigned long flags;

			spin_lock_irqsave(&ip_state_lock, flags);
			if (cmd == IP_NAT_SO_DELETE) 
				ret = delete_rule(&ruleuser, mtype);
			else {
				IP_STATE_ASSERT(cmd == IP_NAT_SO_INSERT);
				ret = insert_rule(&ruleuser, mtype);
			}
			spin_unlock_irqrestore(&ip_state_lock, flags);
		} else {
			DEBUGP("rule_user: no mapping type `%s'\n",
			       ruleuser.typename);
			ret = -EINVAL;
		}
	}

	return ret;
}

static struct nf_setsockopt_ops ip_nat_sockopts
= { NULL, PF_INET, IP_NAT_BASE_CTL, IP_NAT_SO_MAX, do_ip_nat_ctl };

#ifdef CONFIG_PROC_FS
static unsigned int getmask(u_int32_t mask)
{
	int i;

	for (i = 0; mask & 0x80000000; i++) 
		mask <<= 1;

	return i;
}

static unsigned int
dump_rule(char *buffer, const struct ip_nat_rule *rule, unsigned int i)
{
	unsigned int len = 0;
	struct ip_nat_protocol *proto
		= find_nat_proto(rule->match.dst.protonum);

	len += sprintf(buffer + len, "%s [%s] ", 
		       rule->mtype ? rule->mtype->name : "generic",
		       rule->maniptype == IP_NAT_MANIP_SRC ? "SRC" : "DST");
	len += sprintf(buffer + len,
		       "%u.%u.%u.%u/%u->%u.%u.%u.%u/%u ",
		       IP_PARTS(rule->match.src.ip),
		       getmask(rule->mask.src.ip),
		       IP_PARTS(rule->match.dst.ip),
		       getmask(rule->mask.dst.ip));
	if (rule->mask.dst.protonum) {
		len += sprintf(buffer + len, "proto=%u ", 
			       rule->match.dst.protonum);
		len += proto->print(buffer + len, &rule->match, &rule->mask);
	}
	/* Print out range being mapped to. */
	len += sprintf(buffer + len, "TO: ");
	if (rule->mr.range[i].flags & IP_NAT_RANGE_MAP_IPS) {
		len += sprintf(buffer + len, "%u.%u.%u.%u",
			       IP_PARTS(rule->mr.range[i].min_ip));
		if (rule->mr.range[i].max_ip != rule->mr.range[i].min_ip)
			len += sprintf(buffer + len, "-%u.%u.%u.%u",
				       IP_PARTS(rule->mr.range[i].max_ip));
		len += sprintf(buffer + len, " ");
	}
	if (rule->mr.range[i].flags & IP_NAT_RANGE_PROTO_SPECIFIED)
		len += proto->print_range(buffer + len, &rule->mr.range[i]);

	return len;
}

static int 
get_rules(char *buffer, char **start, off_t offset, int length, int dummy)
{
	struct ip_nat_rule *rule;
	off_t upto = 0;
	unsigned long flags;
	int len = 0,last_len = 0;

	spin_lock_irqsave(&ip_state_lock, flags);
	for (rule = nat_rules; rule; rule = rule->next) {
		unsigned int i;

		for (i = 0; i < rule->mr.rangesize; i++, upto++) {
			if (upto < offset) continue;

			last_len = len;
			len += dump_rule(buffer+len, rule, i);
			if (len > length) {
				len = last_len;
				goto out;
			}
		}
	}

 out:
	spin_unlock_irqrestore(&ip_state_lock, flags);

	DEBUGP("nat_get_rules: Length = %i (of %i).  Offset = %li.\n",
	       len, length, upto);
	/* `start' hack - see fs/proc/generic.c line ~165 */
	*start = (char *)((unsigned int)upto - offset);
	return len;
}

struct proc_dir_entry ip_nat_rule_procinfo = {
	0, 12, "ip_nat_rules", S_IFREG | S_IRUGO,
	1, 0, 0, 0, &proc_net_inode_operations, get_rules
};
#endif

int ip_nat_rule_init(void)
{
#ifdef CONFIG_PROC_FS
	proc_net_register(&ip_nat_rule_procinfo);
#endif
	return nf_register_sockopt(&ip_nat_sockopts);
}

void ip_nat_rule_cleanup(void)
{
	nf_unregister_sockopt(&ip_nat_sockopts);

	/* Delete all the rules. */
	while (nat_rules)
		modreg_delete(&nat_rules, nat_rules);

#ifdef CONFIG_PROC_FS
	proc_net_unregister(ip_nat_rule_procinfo.low_ino);
#endif
}
