/* This file contains all the functions required for the standalone
   ip_nat module.

   These are not required by the compatibility layer.
*/

/* (c) 1999 Paul `Rusty' Russell.  Licenced under the GNU General
   Public Licence. */

#define EXPORT_SYMTAB
#include <linux/config.h>
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <net/checksum.h>
#include <asm/spinlock.h>

#include "../state/struct_reserve.h"
#include "ip_nat.h"
#include "ip_nat_rule.h"
#include "ip_nat_protocol.h"
#include "ip_nat_core.h"

static unsigned int 
alloc_null_binding(struct ip_state *state, const struct device *device)
{
	/* Force range to this IP; let proto decide mapping for
	   per-proto parts (hence not IP_NAT_RANGE_PROTO_SPECIFIED). */
	struct ip_nat_multi_range mr = { 1, 
					 { { IP_NAT_RANGE_MAP_IPS, 
					     state->tuplehash[0].tuple.src.ip,
					     state->tuplehash[0].tuple.src.ip,
					     { { 0 } }, { { 0 } } } } };

	return ip_nat_setup_info(state, IP_NAT_MANIP_SRC, &mr, device);
}

static unsigned int 
init_state(struct ip_state *state, const struct device *device)
{
	struct ip_nat_rule *rule;
	struct ip_nat_info *info = (void *)state + ip_nat_offset;
	unsigned int hash = hash_by_src(&state->tuplehash[0].tuple.src,
					state->tuplehash[0].tuple.dst
					.protonum);
	unsigned int ret;

	/* Put it in bysource hash. */
	IP_STATE_ASSERT(!info->bysource.state);
	modreg_prepend(&bysource[hash], &info->bysource);
	info->bysource.state = state;

	rule = find_rule(&state->tuplehash[0].tuple);

	/* If there's a rule which matches, set up the translation. */
	if (rule) {
		if (rule->mtype)
			ret = rule->mtype->setup(state, rule->maniptype,
						 &rule->mr, device);
		else ret = ip_nat_setup_info(state, rule->maniptype,
					     &rule->mr, device);
	} else {
		/* Set up null or pseudo-null binding. */
		ret = alloc_null_binding(state, device);
	}

	if (ret == NF_ACCEPT) {
		unsigned int hash
			= hash_by_ipsproto(state->tuplehash[1].tuple.src.ip,
					   state->tuplehash[1].tuple.dst.ip,
					   state->tuplehash[1]
					   .tuple.dst.protonum);
		info->byipsproto.state = state;
		modreg_prepend(&byipsproto[hash], &info->byipsproto);
		info->initialized = 1;
	}

	return ret;
}

static unsigned int
warn_related(struct ip_state *state, int dir, struct iphdr *iph)
{
	/* FIXME: build infrastructure here. --RR */
	return NF_ACCEPT;
}

static unsigned int
ip_nat(struct sk_buff *skb, const struct device *device, unsigned int hook)
{
	struct ip_state *state = (struct ip_state *)skb->nfmark;
	struct ip_nat_info *info = (void *)state + ip_nat_offset;
	unsigned long flags;
	unsigned int ret, conn_dir = (skb->nfreason >= IPS_ESTABLISHED_REPLY);

	/* FIXME: One day, fill in properly. --RR */
	skb->nfcache |= NFC_UNKNOWN | NFC_ALTERED;

	/* If we had a hardware checksum before, it's now invalid */
	skb->ip_summed = CHECKSUM_NONE;

	spin_lock_irqsave(&ip_state_lock, flags);

	switch (skb->nfreason) {
	case IPS_INVALID:
		ret = NF_ACCEPT;
		goto out;

	case IPS_RELATED_REPLY:  /* Always ICMP */
	case IPS_RELATED:
		if (skb->nh.iph->protocol == IPPROTO_ICMP) {
			icmp_reply_translation(skb, state, conn_dir);
			ret = NF_ACCEPT;
			goto out;
		}
	case IPS_NEW:
		/* Seen it before? */
		if (info->initialized) {
			DEBUGP("Not initializing state\n");
			break;
		}

		DEBUGP("Initializing state\n");
		/* Device == NULL means a local packet, since it the hook
		   is either NF_IP_PRE_ROUTING or NF_IP_LOCAL_IN. */
		ret = init_state(state, device);
		if (ret == NF_ACCEPT) break;
		else goto out;

	case IPS_ESTABLISHED_SPECIFIES_RELATED:
	case IPS_ESTABLISHED_REPLY_SPECIFIES_RELATED:
		ret = warn_related(state, conn_dir, skb->nh.iph);
		if (ret != NF_ACCEPT)
			goto out;
		break;
	}

	if (info->dirinfo[conn_dir].hook == hook)
		do_binding(info, conn_dir, skb->nh.iph, skb->len);

	ret = NF_ACCEPT;

 out:
	spin_unlock_irqrestore(&ip_state_lock, flags);
	return ret;
}

static unsigned int 
ip_nat_fn(unsigned int hooknum,
	  struct sk_buff **pskb,
	  const struct device *in,
	  const struct device *out)
{
	return ip_nat(*pskb, in, hooknum);
}

/* Packet filtering is priority 0: for local_in, we want to be after
   it, for local_out, we want to be before.  And after connection tracking.
*/
static struct nf_hook_ops ip_nat_in_ops = { NULL, ip_nat_fn, NULL,
					   PF_INET, NF_IP_PRE_ROUTING, 0 };
static struct nf_hook_ops ip_nat_out_ops = { NULL, ip_nat_fn, NULL,
					    PF_INET, NF_IP_POST_ROUTING, 0 };
static struct nf_hook_ops ip_nat_local_in_ops = { NULL, ip_nat_fn, NULL,
						 PF_INET, NF_IP_LOCAL_IN,
						  100 };
static struct nf_hook_ops ip_nat_local_out_ops = { NULL, ip_nat_fn, NULL,
						  PF_INET, NF_IP_LOCAL_OUT,
						  -100 };
/* Protocol registration. */
int ip_nat_protocol_register(struct ip_nat_protocol *proto)
{
	if (!modreg_named_insert(&protos, proto))
		return -EEXIST;

	MOD_INC_USE_COUNT;
	return 0;
}

void ip_nat_protocol_unregister(struct ip_nat_protocol *proto)
{
	modreg_delete(&protos, proto);
	MOD_DEC_USE_COUNT;
}

static int init_or_cleanup(int init)
{
	int ret = 0;
	unsigned long flags;

	spin_lock_irqsave(&ip_state_lock, flags);

	if (!init) goto cleanup;

	ret = nf_register_hook(&ip_nat_in_ops);
	if (ret < 0) {
		printk("ip_nat_init: can't register in hook.\n");
		goto cleanup_nothing;
	}
	ret = nf_register_hook(&ip_nat_out_ops);
	if (ret < 0) {
		printk("ip_nat_init: can't register out hook.\n");
		goto cleanup_inops;
	}
	ret = nf_register_hook(&ip_nat_local_in_ops);
	if (ret < 0) {
		printk("ip_nat_init: can't register local in hook.\n");
		goto cleanup_outops;
	}
	ret = nf_register_hook(&ip_nat_local_out_ops);
	if (ret < 0) {
		printk("ip_nat_init: can't register local out hook.\n");
		goto cleanup_localinops;
	}
	ret = ip_nat_rule_init();
	if (ret < 0) {
		printk("ip_nat_init: can't setup rules.\n");
		goto cleanup_localoutops;
	}

	ip_nat_init();
	spin_unlock_irqrestore(&ip_state_lock, flags);
	__MOD_INC_USE_COUNT(ip_state_module);
	return ret;

 cleanup:
	__MOD_DEC_USE_COUNT(ip_state_module);
	ip_nat_cleanup();
	ip_nat_rule_cleanup();
 cleanup_localoutops:
	nf_unregister_hook(&ip_nat_local_out_ops);
 cleanup_localinops:
	nf_unregister_hook(&ip_nat_local_in_ops);
 cleanup_outops:
	nf_unregister_hook(&ip_nat_out_ops);
 cleanup_inops:
	nf_unregister_hook(&ip_nat_in_ops);
 cleanup_nothing:
	spin_unlock_irqrestore(&ip_state_lock, flags);
	return ret;
}

int init_module(void)
{
	return init_or_cleanup(1);
}

void cleanup_module(void)
{
	init_or_cleanup(0);
}

EXPORT_SYMBOL(ip_nat_setup_info);
EXPORT_SYMBOL(ip_nat_register_mapping_type);
EXPORT_SYMBOL(ip_nat_unregister_mapping_type);
