/* Masquerading compatibility layer.

   Note that there are no restrictions on other programs binding to
   ports 61000:65095 (in 2.0 and 2.2 they get EADDRINUSE).  Just DONT
   DO IT.
 */
#include <linux/skbuff.h>
#include <linux/ip.h>
#include <linux/icmp.h>
#include <linux/udp.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netdevice.h>
#include <linux/inetdevice.h>
#include <net/route.h>
#include "../state/ip_state.h"
#include "../state/ip_state_core.h"
#include "../NAT/ip_nat.h"
#include "../NAT/ip_nat_core.h"

unsigned int
do_masquerade(struct sk_buff *skb, const struct device *dev)
{
	struct iphdr *iph = skb->nh.iph;
	struct ip_nat_info *info;
	unsigned int ret;
	unsigned long flags;

	printk("do_masquerade:\n");
	/* Sorry, only ICMP, TCP and UDP. */
	if (iph->protocol != IPPROTO_ICMP
	    && iph->protocol != IPPROTO_TCP
	    && iph->protocol != IPPROTO_UDP)
		return NF_DROP;

	/* Feed it to state tracking; in fact we're in NF_IP_FORWARD,
           but state tracking doesn't expect that */
	ret = ip_state_in(NF_IP_POST_ROUTING, &skb, dev, NULL);
	if (ret != NF_ACCEPT) {
		printk("ip_state_in returned %u.\n", ret);
		return ret;
	}

	if (skb->nfreason == IPS_INVALID) {
		printk("ip_state_in set to invalid state.\n");
		return NF_DROP;
	}

	/* Setup the masquerade, if not already */
	info = (void *)skb->nfmark + ip_nat_offset;
	if (!info->initialized) {
		u_int32_t newsrc;
		struct rtable *rt;
		struct ip_nat_multi_range range;

		/* Pass 0 instead of saddr, since it's going to be changed
		   anyway. */
		if (ip_route_output(&rt, iph->daddr, 0, 0, 0) != 0) {
			DEBUGP("ipnat_rule_masquerade: Can't reroute.\n");
			return NF_DROP;
		}
		newsrc = inet_select_addr(rt->u.dst.dev, rt->rt_gateway,
					  RT_SCOPE_UNIVERSE);
		ip_rt_put(rt);
		range = ((struct ip_nat_multi_range)
			 { 1, 
			   {{IP_NAT_RANGE_MAP_IPS|IP_NAT_RANGE_PROTO_SPECIFIED,
			     newsrc, newsrc, 
			     {{ htons(61000) }}, {{ htons(65095) }} } } });

		spin_lock_irqsave(&ip_state_lock, flags);
		ip_nat_setup_info((struct ip_state *)skb->nfmark,
				  IP_NAT_MANIP_SRC, &range, dev);
		info->initialized = 1;
		spin_unlock_irqrestore(&ip_state_lock, flags);
	} else
		printk("Masquerading already done on this conn.\n");

	spin_lock_irqsave(&ip_state_lock, flags);
	do_binding(info, (skb->nfreason >= IPS_ESTABLISHED_REPLY),
		   skb->nh.iph, skb->len);
	spin_unlock_irqrestore(&ip_state_lock, flags);
	return NF_ACCEPT;
}

void
check_for_demasq(struct sk_buff *skb)
{
	struct ip_state_tuple tuple;
	struct iphdr *iph = skb->nh.iph;
	unsigned long flags;
	struct ip_state_protocol *protocol;
	struct ip_state_tuple_hash *h;

	printk("check_for_demasq\n");
	spin_lock_irqsave(&ip_state_lock, flags);

	protocol = find_proto(iph->protocol);
	/* We don't feed packets to state system unless we know
           they're part of an connection already established by an
           explicit masq command. */
	switch (iph->protocol) {
	case IPPROTO_ICMP:
		/* ICMP errors. */ 
		if (icmp_error_track(skb)) {
			if (skb->nfreason != IPS_INVALID)
				icmp_reply_translation(skb, 
						       (struct ip_state *)
						       skb->nfmark,
						       skb->nfreason
						       == IPS_RELATED_REPLY);
			spin_unlock_irqrestore(&ip_state_lock, flags);
			return;
		}
		/* Fall thru... */
	case IPPROTO_TCP:
	case IPPROTO_UDP:
		if (!get_tuple(iph, skb->len, &tuple, protocol)) {
			spin_unlock_irqrestore(&ip_state_lock, flags);
			printk("ip_fw_compat_masq: Couldn't get tuple\n");
			return;
		}
		break;

	default:
		/* Not ours... */
		spin_unlock_irqrestore(&ip_state_lock, flags);
		return;
	}

	h = ip_state_find(&tuple, NULL);

	/* MUST be found, and MUST be reply. */
	if (h && h == &h->state->tuplehash[1]) {
		/* Feed to state system. */
		if (ip_state_in(NF_IP_PRE_ROUTING, &skb, NULL, NULL)
		    == NF_ACCEPT
		    && skb->nfreason != IPS_INVALID) {
			struct ip_nat_info *info 
				= (void *)skb->nfmark + ip_nat_offset;

			do_binding(info,
				   (skb->nfreason >= IPS_ESTABLISHED_REPLY),
				   iph, skb->len);
		} else
			printk("ip_fw_compat_masq: state didn't like\n");
	} else {
		printk("ip_fw_compat_masq: no matching state\n");
	}
		

	spin_unlock_irqrestore(&ip_state_lock, flags);
	return;
}

int ip_fw_masq_timeouts(void *user, int len)
{
	/* FIXME: Do somthing! --RR */
	return 0;
}

void masq_init(void)
{
	ip_state_init();
	ip_nat_init();
}

void masq_cleanup(void)
{
	ip_nat_cleanup();
	ip_state_cleanup();
}
