/* Connection state tracking for netfilter.  This is separated from,
   but required by, the NAT layer; it can also be used by an iptables
   extention. */

/* (c) 1999 Paul `Rusty' Russell.  Licenced under the GNU General
   Public Licence. */

#define __NO_VERSION__
#include <linux/config.h>
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <net/checksum.h>
#include <asm/spinlock.h>

#include "struct_reserve.h"
#include "ip_state.h"
#include "ip_state_protocol.h"
#include "ip_state_expect.h"
#include "ip_state_core.h"

/* If we're the first tuple, it's the original dir. */
#define DIRECTION(h) (&(h)->state->tuplehash[1] == (h))

/* One big spinlock. FIXME: Get funky --RR */
#ifdef CONFIG_SMP
static spinlock_t ip_state_lock = SPIN_LOCK_UNLOCKED;
#endif

/* If we're the first tuple, it's the original dir. */
#define DIRECTION(h) (&(h)->state->tuplehash[1] == (h))

void (*ip_state_destroyed)(struct ip_state *state) = NULL;
static struct ip_state_expect *expect_list = NULL;
struct ip_state_protocol *protocol_list;
static struct ip_state_tuple_hash *ip_state_hash[IP_STATE_HTABLE_SIZE] = { NULL };

extern struct ip_state_protocol ip_state_generic_protocol;

DECLARE_STRUCT_RESERVABLE(ip_state);

static int cmpfn(const void *arg, const void *curr)
{
	/* We cast to unsigned long because gcc bitches if we cast to
           u_int8_t */
	return (unsigned long)arg == ((struct ip_state_protocol *)curr)->proto;
}

struct ip_state_protocol *find_proto(u_int8_t protocol)
{
	struct ip_state_protocol *p
		= modreg_find(&protocol_list, (void *)(unsigned long)protocol,
			      &cmpfn);

	if (!p)
		p = &ip_state_generic_protocol;
	return p;
}

static inline u_int32_t 
hash_state(const struct ip_state_tuple *tuple)
{
#if 0
	dump_tuple(tuple);
#endif
#ifdef DEBUG_STATE
	if (tuple->src.pad)
		DEBUGP("Tuple %p has non-zero padding.\n", tuple);
#endif
	/* ntohl because more differences in low bits. */
	/* To ensure that halves of the same connection don't hash
	   clash, we bitshift the source per-proto. */
	return ntohl(tuple->src.ip + tuple->dst.ip
		     + (tuple->src.u.all<<1) + tuple->dst.u.all
		     + tuple->dst.protonum)
		% IP_STATE_HTABLE_SIZE;
}

inline int 
get_tuple(const struct iphdr *iph, size_t len, struct ip_state_tuple *tuple,
	  struct ip_state_protocol *protocol)
{
	int ret;

	/* FIXME:  Be clever.  --RR */
	if (ntohs(iph->frag_off) & IP_OFFSET) {
		static int warned = 0;
		if (!warned)
			printk("ip_state_core: Frag of proto %u.\n", 
			       iph->protocol);
		warned = 1;
		return 0;
	}
	else if (iph->ihl * 4 + protocol->min_length > len)
		return 0;

	tuple->src.ip = iph->saddr;
	tuple->src.pad = 0;
	tuple->dst.ip = iph->daddr;
	tuple->dst.protonum = iph->protocol;

	ret = protocol->pkt_to_tuple((u_int32_t *)iph + iph->ihl,
				     len - 4*iph->ihl,
				     tuple);
	return ret;
}

/* Simple state table: indexed by status and direction. */
static struct 
{
	/* new status. */
	enum ip_state_status status;

	/* state type for skb. */
	enum ip_statetype statetype;
} states[3][2] = {
	/* IPS_AWAITING_REPLY */
	{ { IPS_AWAITING_REPLY, IPS_NEW_REPEAT },
	  { IPS_BOTH_WAYS, IPS_ESTABLISHED_REPLY } },

	/* IPS_BOTH_WAYS */
	{ { IPS_BOTH_WAYS, IPS_ESTABLISHED },
	  { IPS_BOTH_WAYS, IPS_ESTABLISHED_REPLY } },

	/* IPS_EXPECTED_AWAITING_REPLY */
	{ { IPS_EXPECTED_AWAITING_REPLY, IPS_RELATED_REPEAT },
	  { IPS_BOTH_WAYS, IPS_ESTABLISHED_REPLY } }
};

/* find a connection corresponding to a tuple. */
static struct ip_state_tuple_hash **
find_state(const struct ip_state_tuple *tuple)
{
	struct ip_state_tuple_hash **i;

#if 0
	DEBUGP("Searching for tuple: "); dump_tuple(tuple);
#endif
	for (i = &ip_state_hash[hash_state(tuple)]; *i; i = &(*i)->next)
		if (memcmp(tuple, &(*i)->tuple, sizeof(*tuple)) == 0)
			break;
		else {
#if 0
			DEBUGP("Failed compare against: ");
			dump_tuple(&(*i)->tuple);
#endif
		}

	return i;
}

static int
invert_tuple(struct ip_state_tuple *inverse,
	     const struct ip_state_tuple *orig,
	     const struct ip_state_protocol *protocol)
{
	inverse->src.ip = orig->dst.ip;
	inverse->src.pad = 0;
	inverse->dst.ip = orig->src.ip;
	inverse->dst.protonum = orig->dst.protonum;

	return protocol->invert_tuple(inverse, orig);
}

static void free_state(struct ip_state *state)
{
	DEBUGP("Freeing state %p: ", state);
	if (state->timeout.function) {
		DEBUGP("deleting timeout.\n");
		del_timer(&state->timeout);
	}
	else DEBUGP("not deleting timeout.\n");
	find_proto(state->tuplehash[0].tuple.dst.protonum)->destroy(state);
	if (ip_state_destroyed) 
		ip_state_destroyed(state);

	kfree(state);
}

/* Delete a state. */
void ip_state_delete(struct ip_state *state)
{
	struct ip_state_tuple_hash **hptr;

#if 0
	STATE_ASSERT(ip_state_lock.lock);
#endif

#if 0
	if (!state || !check_state(state)) {
		DEBUGP("ip_state_delete: state %p DNE.\n", state);
		return;
	}
#endif

	if (state->tuplehash[0].state) {
		DEBUGP("ip_state_delete: state %p dir 0.\n", state);
		hptr = find_state(&state->tuplehash[0].tuple);
#ifdef DEBUG_STATE
		if (!*hptr) {
			DEBUGP("ip_state_delete: state %p dir 0 DNE.\n",
			       state);
		}
		else
#endif
		*hptr = (*hptr)->next;
	}
	if (state->tuplehash[1].state) {
		DEBUGP("ip_state_delete: state %p dir 1.\n", state);
		hptr = find_state(&state->tuplehash[1].tuple);
#ifdef DEBUG_STATE
		if (!*hptr) {
			DEBUGP("ip_state_delete: state %p dir 1 DNE.\n",
			       state);
		}
		else
#endif
		*hptr = (*hptr)->next;
	}

	free_state(state);
}

static void ip_state_delete_timeout(unsigned long ul_state)
{
	unsigned long flags;

	DEBUGP("ip_state_delete_timeout(%p)\n", (struct ip_state *)ul_state);
	spin_lock_irqsave(&ip_state_lock, flags);
	ip_state_delete((struct ip_state *)ul_state);
	spin_unlock_irqrestore(&ip_state_lock, flags);
}

/* Compare all but src per-proto part. */
static int expect_cmp(const void *arg, const void *curr)
{
	const struct ip_state_tuple *tuple = arg;
	const struct ip_state_expect *expect = curr;

	return (tuple->src.ip == expect->tuple.src.ip
		&& tuple->dst.ip == expect->tuple.dst.ip
		&& tuple->dst.u.all == expect->tuple.dst.u.all
		&& tuple->dst.protonum == expect->tuple.dst.protonum);
}

/* Returns TRUE if it dealt with the skb. */
int icmp_error_track(struct sk_buff *skb)
{
	struct iphdr *iph = skb->nh.iph;
	struct icmphdr *hdr = (struct icmphdr *)((u_int32_t *)iph + iph->ihl);
	struct ip_state_tuple innertuple, origtuple;
	struct iphdr *inner = (struct iphdr *)(hdr + 1);
	size_t datalen = skb->len - iph->ihl*4 - sizeof(*hdr);
	struct ip_state_tuple_hash *h;
	struct ip_state_protocol *innerproto;

	if (skb->len < iph->ihl * 4 + sizeof(struct icmphdr)) {
		DEBUGP("icmp_error_track: too short\n");
		goto inval;
	}

	if (hdr->type != ICMP_DEST_UNREACH
	    && hdr->type != ICMP_SOURCE_QUENCH
	    && hdr->type != ICMP_TIME_EXCEEDED
	    && hdr->type != ICMP_PARAMETERPROB
	    && hdr->type != ICMP_REDIRECT)
		return 0;

	/* Ignore it if the checksum's bogus. */
	if (ip_compute_csum((unsigned char *)hdr, sizeof(*hdr) + datalen)) {
		DEBUGP("icmp_error_track: bad csum\n");
		goto inval;
	}

	innerproto = find_proto(inner->protocol);
	/* Are they talking about one of our connections? */
	if (inner->ihl * 4 + innerproto->min_length > datalen
	    || !get_tuple(inner, datalen, &origtuple, innerproto)) {
		DEBUGP("icmp_error_track: Failed to get tuple\n");
		goto inval;
	}

	/* Ordinarily, we'd expect the inverted tupleproto, but it's
	   been preserved inside the ICMP. */
	if (!invert_tuple(&innertuple, &origtuple, innerproto)) {
		DEBUGP("icmp_error_track: Can't invert tuple\n");
		goto inval;
	}
	h = *find_state(&innertuple);
	if (!h) {
		DEBUGP("icmp_error_track: no match\n");
		goto inval;
	}

	DEBUGP("icmp_error_track: Found tuple.\n");
	if (DIRECTION(h))
		skb->nfreason = IPS_RELATED_REPLY;
	else
		skb->nfreason = IPS_RELATED;
	skb->nfmark = (unsigned long)h->state;
	return 1;

 inval:
	skb->nfreason = IPS_INVALID;
	skb->nfmark = (unsigned long)NULL;

	return 1;
}

/* Netfilter hook itself. */
unsigned int ip_state_in(unsigned int hooknum,
			 struct sk_buff **pskb,
			 const struct device *in,
			 const struct device *out)
{
	unsigned int ret;
	struct ip_state_tuple tuple;
	struct ip_state_tuple_hash **hptr;
	struct iphdr *iph = (*pskb)->nh.iph;
	unsigned long flags;
	struct ip_state_protocol *protocol;

	/* FIXME: Do this right please. --RR */
	(*pskb)->nfcache |= NFC_UNKNOWN;

	/* Deja vu?  Loopback. */
	if ((*pskb)->nfreason)
		return NF_ACCEPT;

	spin_lock_irqsave(&ip_state_lock, flags);

	if (iph->protocol == IPPROTO_ICMP && icmp_error_track(*pskb)) {
		ret = NF_ACCEPT;
		goto end;
	}

	protocol = find_proto(iph->protocol);
	if (!get_tuple(iph, (*pskb)->len, &tuple, protocol)) {
		(*pskb)->nfreason = IPS_INVALID;
		(*pskb)->nfmark = (unsigned long)NULL;

		ret = NF_ACCEPT;
		goto end;
	}

	hptr = find_state(&tuple);
	if (*hptr) {
		/* It exists! */
		int dir = DIRECTION(*hptr);
		enum ip_statetype statetype;

#if 1
		DEBUGP("We have hptr %p for tuple (status = %u, dir = %u).\n",
		       *hptr, (*hptr)->state->status, dir);
#endif
		statetype = states[(*hptr)->state->status][dir].statetype;
		(*hptr)->state->status
			= states[(*hptr)->state->status][dir].status;

		/* Established may delete; don't reference hptr after. */
		(*pskb)->nfmark = (unsigned long)(*hptr)->state;
		ret = protocol->established((*hptr)->state, iph, 
					    (*pskb)->len, dir,
					    &statetype);
		(*pskb)->nfreason = statetype;
	} else {
		/* Allocate a new state. */
		struct ip_state *state;
		struct ip_state_tuple repl_tuple;
		enum ip_statetype statetype = IPS_NEW;
		size_t hash, repl_hash;
		struct ip_state_expect *expected;

#if 0
		DEBUGP("No hptr for tuple [hash %u]: ", hash_state(&tuple));
		dump_tuple(&tuple);
#endif
		if (!invert_tuple(&repl_tuple, &tuple, protocol)) {	
			DEBUGP("Can't invert tuple.\n");
			ret = NF_DROP;
			goto end;
		}

		state = STRUCT_RESERVE_ALLOCATE(ip_state, GFP_ATOMIC);
		if (!state) {
			DEBUGP("Can't allocate state.\n");
			ret = NF_DROP;
			goto end;
		}

		hash = hash_state(&tuple);
		repl_hash = hash_state(&repl_tuple);

		DEBUGP("Hash for tuple = %u.  Reply = %u.\n", hash, repl_hash);
		*state = ((struct ip_state)
			  {{ { NULL, tuple, state },
			     { NULL, repl_tuple, state } },
			   IPS_AWAITING_REPLY, 
			   { NULL, NULL, jiffies + protocol->expiry, 
			     (unsigned long)state, NULL },
			   NULL });
		/* Sew in at head of hash list. */
		modreg_prepend(&ip_state_hash[hash], &state->tuplehash[0]);
		modreg_prepend(&ip_state_hash[repl_hash],&state->tuplehash[1]);

		if (protocol->expiry) {
			init_timer(&state->timeout);
			state->timeout.function	= ip_state_delete_timeout;
			add_timer(&state->timeout);
		}

		expected = modreg_find(&expect_list, &tuple, &expect_cmp);
		if (expected) {
			/* Welcome, Mr. Bond.  We've been expecting you... */
			DEBUGP("Made related state %p.\n", state);
			state->status = IPS_EXPECTED_AWAITING_REPLY;
			statetype = IPS_RELATED;
			modreg_delete(&expect_list, expected);
		}
		else DEBUGP("Made state %p.\n", state);

		(*pskb)->nfmark = (unsigned long)state;
		ret = protocol->new(state, iph, (*pskb)->len, &statetype);
		(*pskb)->nfreason = statetype;
	}
			
 end:
	spin_unlock_irqrestore(&ip_state_lock, flags);
	return ret;
}

/* Find a connection corresponding to a tuple. */
struct ip_state_tuple_hash *
ip_state_find(const struct ip_state_tuple *tuple,
	      const struct ip_state *ignored_state)
{
	struct ip_state_tuple_hash *i;

#if 0
	IP_STATE_ASSERT(ip_state_lock.lock);
#endif

	for (i = ip_state_hash[hash_state(tuple)]; i; i = i->next)
		if (i->state != ignored_state
		    && memcmp(tuple, &i->tuple, sizeof(*tuple)) == 0)
			break;

	return i;
}

#ifdef DEBUG_STATE
static void dump_tuple(const struct ip_state_tuple *tuple)
{
	printk("tuple %p: %u %u.%u.%u.%u:%u -> %u.%u.%u.%u:%u\n",
	       tuple, tuple->dst.protonum,
	       IP_PARTS(tuple->src.ip), ntohs(tuple->src.u.tcp.port),
	       IP_PARTS(tuple->dst.ip), ntohs(tuple->dst.u.tcp.port));
}
#endif

int invert_tuplepr(struct ip_state_tuple *inverse,
		   const struct ip_state_tuple *orig)
{
	return invert_tuple(inverse, orig, find_proto(orig->dst.protonum));
}

static void
add_hash(struct ip_state_tuple_hash *hash, struct ip_state *state)
{
	unsigned int index = hash_state(&hash->tuple);
	hash->state = state;
	hash->next = ip_state_hash[index];
	ip_state_hash[index] = hash;
}

/* Alter reply tuple. */
void ip_state_alter_reply(struct ip_state *state, 
			  const struct ip_state_tuple *newreply)
{
	size_t hash = hash_state(&state->tuplehash[1].tuple);
#if 0
	IP_STATE_ASSERT(ip_state_lock.lock);
#endif

	state->tuplehash[1].tuple = *newreply;

	/* If it's not hashed, don't rehash. */
	if (state->tuplehash[1].state) {
		modreg_delete(&ip_state_hash[hash], &state->tuplehash[1]);
		add_hash(&state->tuplehash[1], state);
	}
}

#ifdef DEBUG_STATE
static int check_state(struct ip_state *state)
{
	/* Traverse hash looking for pointer to state. */
	unsigned int i;

	for (i = 0; i < IP_STATE_HTABLE_SIZE; i++) {
		struct ip_state_tuple_hash *h;

		for (h = ip_state_hash[i]; h; h = h->next) {
			if (h->state == state)
				return 1;
		}
	}

	return 0;
}
#endif

/* Expire in this direction.  If expired in both dirs, delete. */
void ip_state_delete_maybe(struct ip_state *state, int direction)
{
	struct ip_state_tuple_hash **hptr;

#ifdef DEBUG_STATE
	if (!state || !check_state(state)) {
		DEBUGP("ip_state_delete_maybe: state %p DNE.\n", state);
		return;
	}
#endif

	state->tuplehash[direction].state = NULL;
	hptr = find_state(&state->tuplehash[direction].tuple);
	*hptr = (*hptr)->next;

	if (!state->tuplehash[!direction].state)
		free_state(state);
}

/* Delete all the states. */
void ip_state_cleanup(void)
{
	unsigned int i;

	/* Delete them all. */
	for (i = 0; i < IP_STATE_HTABLE_SIZE; i++) {
		struct ip_state_tuple_hash *h;

		for (h = ip_state_hash[i]; h; h = h->next) {
			/* Only count originals; unless original not hashed */
			if (!DIRECTION(h) && h->state->tuplehash[0].state)
				continue;

			free_state(h->state);
		}
	}
}

void ip_state_init(void)
{
	/* Sew in builtin protocols. */
	protocol_list = &ip_state_protocol_tcp;
	ip_state_protocol_tcp.next = &ip_state_protocol_udp;
	ip_state_protocol_udp.next = &ip_state_protocol_icmp;

	ip_state_protocol_tcp_init();
}
