/* This file contains all the functions required for the standalone
   ip_state module.

   These are not required by the compatibility layer.
*/

/* (c) 1999 Paul `Rusty' Russell.  Licenced under the GNU General
   Public Licence. */

#define EXPORT_SYMTAB
#include <linux/config.h>
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/netfilter.h>
#include <linux/netfilter_ipv4.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/proc_fs.h>
#include <net/checksum.h>
#include <asm/spinlock.h>

#include "struct_reserve.h"
#include "ip_state.h"
#include "ip_state_protocol.h"
#include "ip_state_expect.h"
#include "ip_state_core.h"

struct module *ip_state_module = &__this_module;

static unsigned int 
print_tuple(char *buffer, const struct ip_state_tuple *tuple,
	    struct ip_state_protocol *proto)
{
	int len;

	len = sprintf(buffer, "src=%u.%u.%u.%u dst=%u.%u.%u.%u ",
		      IP_PARTS(tuple->src.ip), IP_PARTS(tuple->dst.ip));

	len += proto->print_tuple(buffer + len, tuple);

	return len;
}

static unsigned int
print_state(char *buffer, const struct ip_state *state)
{
	unsigned int len;
	struct ip_state_protocol *proto
		= find_proto(state->tuplehash[0].tuple.dst.protonum);

	len = sprintf(buffer, "%-8s %u ", 
		      proto->name, state->tuplehash[0].tuple.dst.protonum);

	len += proto->print_state(buffer + len, state);
	len += print_tuple(buffer + len, &state->tuplehash[0].tuple, proto);
	if (state->status == IPS_AWAITING_REPLY 
	    || state->status == IPS_EXPECTED_AWAITING_REPLY)
		len += sprintf(buffer + len, "[UNREPLIED] ");
	len += print_tuple(buffer + len, &state->tuplehash[1].tuple, proto);
	len += sprintf(buffer + len, "\n");

	return len;
}

/* FIXME: Don't print source proto part. --RR */
static unsigned int 
print_expect(char *buffer, const struct ip_state_expect *expect)
{
	unsigned int len;

	len = sprintf(buffer, "EXPECTING: ");
	len += print_tuple(buffer + len, &expect->tuple,
			   find_proto(expect->tuple.dst.protonum));
	len += sprintf(buffer + len, "\n");
	return len;
}

/* If we're the first tuple, it's the original dir. */
#define DIRECTION(h) (&(h)->state->tuplehash[1] == (h))

static int 
list_states(char *buffer, char **start, off_t offset, int length, int dummy)
{
	unsigned int i;
	int len = 0,last_len = 0;
	off_t upto = 0;
	struct ip_state_expect *expect;
	unsigned long flags;

	spin_lock_irqsave(&ip_state_lock, flags);

	/* Traverse hash; print originals then reply. */
	for (i = 0; i < IP_STATE_HTABLE_SIZE; i++) {
		struct ip_state_tuple_hash *h;

		for (h = ip_state_hash[i]; h; h = h->next) {

			IP_STATE_ASSERT(h->state);
			/* Only count originals; unless original not hashed */
			if (!DIRECTION(h) && h->state->tuplehash[0].state)
				continue;

			if (upto++ < offset) continue;

			last_len = len;
			len += print_state(buffer + len, h->state);
			if (len > length) {
				len = last_len;
				goto finished;
			}
		}
	}

	/* Now iterate through expecteds. */
	for (expect = expect_list; expect; expect = expect->next) {
		if (upto++ < offset) continue;

		last_len = len;
		len += print_expect(buffer + len, expect);
		if (len > length) {
			len = last_len;
			goto finished;
		}
	}

 finished:
	spin_unlock_irqrestore(&ip_state_lock, flags);

	/* `start' hack - see fs/proc/generic.c line ~165 */
	*start = (char *)((unsigned int)upto - offset);
	return len;
}

/* State tracking may drop packets, but never alters them, so make
   it the first hook. */
static struct nf_hook_ops ip_state_in_ops 
= { NULL, ip_state_in, NULL, PF_INET, NF_IP_PRE_ROUTING, -200 };
static struct nf_hook_ops ip_state_local_out_ops 
= { NULL, ip_state_in, NULL, PF_INET, NF_IP_LOCAL_OUT, -200 };

#ifdef CONFIG_PROC_FS
struct proc_dir_entry ip_state_procinfo = {
	0, 8, "ip_state", S_IFREG | S_IRUGO,
	1, 0, 0, 0, &proc_net_inode_operations, &list_states
};
#endif

extern struct module *ip_defrag_module;

static int init_or_cleanup(int init)
{
	int ret = 0;
	unsigned long flags;

	spin_lock_irqsave(&ip_state_lock, flags);
	
	if (!init) goto cleanup;

#ifdef CONFIG_PROC_FS
	proc_net_register(&ip_state_procinfo);
#endif
	ret = nf_register_hook(&ip_state_in_ops);
	if (ret < 0) {
		printk("ip_state_init: can't register in hook.\n");
		goto cleanup_nothing;
	}
	ret = nf_register_hook(&ip_state_local_out_ops);
	if (ret < 0) {
		printk("ip_state_init: can't register local out hook.\n");
		goto cleanup_inops;
	}

	ip_state_init();
	spin_unlock_irqrestore(&ip_state_lock, flags);
	__MOD_INC_USE_COUNT(ip_defrag_module);
	return ret;

 cleanup:
	ip_state_cleanup();
	nf_unregister_hook(&ip_state_local_out_ops);
	__MOD_DEC_USE_COUNT(ip_defrag_module);
	
 cleanup_inops:
	nf_unregister_hook(&ip_state_in_ops);
 cleanup_nothing:
#ifdef CONFIG_PROC_FS
	proc_net_unregister(ip_state_procinfo.low_ino);
#endif
	spin_unlock_irqrestore(&ip_state_lock, flags);
	return ret;
}

/* FIXME: Allow NULL functions and sub in pointers to generic for
   them. --RR */
int ip_state_register_protocol(struct ip_state_protocol *proto)
{
	MOD_INC_USE_COUNT;
	/* Not unknown protocol? */
	if (find_proto(proto->proto)->proto != 0)
		return 0;
	modreg_prepend(&protocol_list, proto);
	return 1;
}

void ip_state_unregister_protocol(struct ip_state_protocol *proto)
{
	unsigned int i;
	unsigned long flags;
	
	modreg_delete(&protocol_list, proto);

	spin_lock_irqsave(&ip_state_lock, flags);

	/* Delete those using us.  They'll fall back to generic. */
	/* Traverse hash. */
	for (i = 0; i < IP_STATE_HTABLE_SIZE; i++) {
		struct ip_state_tuple_hash *h;

		for (h = ip_state_hash[i]; h; ) {
			struct ip_state_tuple_hash *next = h->next;

			if (h->tuple.dst.protonum == proto->proto)
				ip_state_delete_maybe(h->state, DIRECTION(h));

			h = next;
		}
	}
	spin_unlock_irqrestore(&ip_state_lock, flags);
	MOD_DEC_USE_COUNT;
}

int init_module(void)
{
	return init_or_cleanup(1);
}

void cleanup_module(void)
{
	init_or_cleanup(0);
}

EXPORT_SYMBOL(ip_state_register_protocol);
EXPORT_SYMBOL(ip_state_unregister_protocol);
EXPORT_SYMBOL(expect_list);
EXPORT_SYMBOL(invert_tuplepr);
EXPORT_SYMBOL(ip_state_alter_reply);
EXPORT_SYMBOL(ip_state_destroyed);
EXPORT_SYMBOL(ip_state_find);
EXPORT_SYMBOL(ip_state_module);
EXPORT_STRUCT_RESERVABLE(ip_state);

/* From protocols/ip_state_tcp.c */
struct ip_state_tcp_helper;
extern int ip_state_tcp_helper_register(struct ip_state_tcp_helper *);
extern void ip_state_tcp_helper_unregister(struct ip_state_tcp_helper *);
EXPORT_SYMBOL(ip_state_tcp_helper_register);
EXPORT_SYMBOL(ip_state_tcp_helper_unregister);

