#define __NO_VERSION__
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/netfilter.h>
#include <linux/module.h>
#include <linux/ip.h>
#include <linux/tcp.h>
#include "../ip_state_protocol.h"
#include "ip_state_tcp.h"
#include "../ip_state_expect.h"

static size_t protocol_offset;
static struct ip_state_tcp_helper *helpers = NULL;

/* FIXME: Examine ipfilter's timeouts and state transitions more
   closely.  They're more complex. --RR */

/* Actually, I believe that neither ipmasq (where this code is stolen
   from) nor ipfilter do it exactly right.  A new state machine taking
   into account packet loss (which creates uncertainty as to exactly
   the state of the connection) is required.  RSN.  --RR */
enum tcp_state {
	TCP_STATE_NONE,
	TCP_STATE_ESTABLISHED,
	TCP_STATE_SYN_SENT,
	TCP_STATE_SYN_RECV,
	TCP_STATE_FIN_WAIT,
	TCP_STATE_TIME_WAIT,
	TCP_STATE_CLOSE,
	TCP_STATE_CLOSE_WAIT,
	TCP_STATE_LAST_ACK,
	TCP_STATE_LISTEN,
	TCP_STATE_MAX
};

static const char *tcp_state_names[] = {
	"NONE",
	"ESTABLISHED",
	"SYN_SENT",
	"SYN_RECV",
	"FIN_WAIT",
	"TIME_WAIT",
	"CLOSE",
	"CLOSE_WAIT",
	"LAST_ACK",
	"LISTEN"
};

#define SECS *HZ
#define MINS * 60 SECS
#define HOURS * 60 MINS
#define DAYS * 24 HOURS

static unsigned long tcp_timeouts[]
= { 30 MINS, 	/*	TCP_STATE_NONE,	*/
    5 DAYS,	/*	TCP_STATE_ESTABLISHED,	*/
    2 MINS,	/*	TCP_STATE_SYN_SENT,	*/
    60 SECS,	/*	TCP_STATE_SYN_RECV,	*/
    2 MINS,	/*	TCP_STATE_FIN_WAIT,	*/
    2 MINS,	/*	TCP_STATE_TIME_WAIT,	*/
    10 SECS,	/*	TCP_STATE_CLOSE,	*/
    60 SECS,	/*	TCP_STATE_CLOSE_WAIT,	*/
    30 SECS,	/*	TCP_STATE_LAST_ACK,	*/
    2 MINS,	/*	TCP_STATE_LISTEN,	*/
};

#define sNO TCP_STATE_NONE
#define sES TCP_STATE_ESTABLISHED
#define sSS TCP_STATE_SYN_SENT
#define sSR TCP_STATE_SYN_RECV
#define sFW TCP_STATE_FIN_WAIT
#define sTW TCP_STATE_TIME_WAIT
#define sCL TCP_STATE_CLOSE
#define sCW TCP_STATE_CLOSE_WAIT
#define sLA TCP_STATE_LAST_ACK
#define sLI TCP_STATE_LISTEN
#define sIV TCP_STATE_MAX

static enum tcp_state tcp_states[2][5][TCP_STATE_MAX] = {
	{
/*	ORIGINAL */
/* 	  sNO, sES, sSS, sSR, sFW, sTW, sCL, sCW, sLA, sLI 	*/
/*syn*/	{sSS, sES, sSS, sES, sSS, sSS, sSS, sSS, sSS, sLI },
/*fin*/	{sTW, sFW, sSS, sTW, sFW, sTW, sCL, sTW, sLA, sLI },
/*ack*/	{sES, sES, sSS, sSR, sFW, sTW, sCL, sCW, sLA, sES },
/*rst*/ {sCL, sCL, sSS, sCL, sCL, sTW, sCL, sCL, sCL, sCL },
/*none*/{sIV, sIV, sIV, sIV, sIV, sIV, sIV, sIV, sIV, sIV }
	},
	{
/*	REPLY */
/* 	  sNO, sES, sSS, sSR, sFW, sTW, sCL, sCW, sLA, sLI 	*/
/*syn*/	{sSR, sES, sES, sSR, sSR, sSR, sSR, sSR, sSR, sSR },
/*fin*/	{sCL, sCW, sSS, sTW, sTW, sTW, sCL, sCW, sLA, sLI },
/*ack*/	{sCL, sES, sSS, sSR, sFW, sTW, sCL, sCW, sCL, sLI },
/*rst*/ {sCL, sCL, sCL, sSR, sCL, sCL, sCL, sCL, sLA, sLI },
/*none*/{sIV, sIV, sIV, sIV, sIV, sIV, sIV, sIV, sIV, sIV }
	}
};

struct protocol_info
{
	enum tcp_state tcp_state;
	struct ip_state_expect expect;
};

static int tcp_pkt_to_tuple(const void *datah, size_t datalen,
			    struct ip_state_tuple *tuple)
{
	const struct tcphdr *hdr = datah;

	tuple->src.u.tcp.port = hdr->source;
	tuple->dst.u.tcp.port = hdr->dest;

	return 1;
}

static int tcp_invert_tuple(struct ip_state_tuple *tuple,
			    const struct ip_state_tuple *orig)
{
	tuple->src.u.tcp.port = orig->dst.u.tcp.port;
	tuple->dst.u.tcp.port = orig->src.u.tcp.port;
	return 1;
}

/* Print out the per-protocol part of the tuple. */
static unsigned int tcp_print_tuple(char *buffer, 
				    const struct ip_state_tuple *tuple)
{
	return sprintf(buffer, "sport=%hu dport=%hu ",
		       ntohs(tuple->src.u.tcp.port),
		       ntohs(tuple->dst.u.tcp.port));
}

/* Print out the private part of the state. */
static unsigned int tcp_print_state(char *buffer, const struct ip_state *state)
{
	struct protocol_info *info = (void *)state + protocol_offset;

	return sprintf(buffer, "%s ", tcp_state_names[info->tcp_state]);
}

static void tcp_destroy(struct ip_state *state)
{
	/* If our expected is in the list, take it out. */
	struct protocol_info *info = (void *)state + protocol_offset;

	if (modreg_inlist(&expect_list, &info->expect))
		modreg_delete(&expect_list, &info->expect);
}

static unsigned int get_state_index(const struct tcphdr *tcph)
{
	if (tcph->rst) return 3;
	else if (tcph->syn) return 0;
	else if (tcph->fin) return 1;
	else if (tcph->ack) return 2;
	else return 4;
}

static int portcmp(const void *arg, const void *curr)
{
	return (unsigned long)arg
		== ((struct ip_state_tcp_helper *)curr)->port;
}

/* Returns verdict for packet, and may modify statetype */
static unsigned int tcp_established(struct ip_state *state,
				    struct iphdr *iph, size_t len,
				    int dir, enum ip_statetype *statetype)
{
	struct protocol_info *info = (void *)state + protocol_offset;
	enum tcp_state newstate;
	struct tcphdr *tcph = (struct tcphdr *)((u_int32_t *)iph + iph->ihl);
	struct ip_state_tcp_helper *helper;

	if (len < (iph->ihl + tcph->doff) *4 ) {
		DEBUGP("ip_state_tcp: Truncated packet.\n");
		*statetype = IPS_INVALID;
		return NF_ACCEPT;
	}

	newstate = tcp_states[dir][get_state_index(tcph)][info->tcp_state];

	/* Invalid */
	if (newstate == TCP_STATE_MAX) {
		DEBUGP("ip_state_tcp: Invalid dir=%i index=%u state=%u\n",
		       dir, get_state_index(tcph), info->tcp_state);
		*statetype = IPS_INVALID;
		return NF_ACCEPT;
	}

	/* Refresh. */
	info->tcp_state = newstate;
	mod_timer(&state->timeout, jiffies + tcp_timeouts[info->tcp_state]);

	/* FIXME: Can cache this for speed. --RR */
	helper = modreg_find(&helpers, 
			     (void *)(dir ? tcph->source : tcph->dest),
			     &portcmp);

	/* Not established yet?  Don't open related! */
	if (helper && state->status == IPS_BOTH_WAYS)
		return helper->established(tcph, len - iph->ihl * 4, dir,
					   state, statetype, &info->expect);
	else return NF_ACCEPT;
}

/* Called when a new connection for this protocol found. */
static unsigned int tcp_new(struct ip_state *state,
			    struct iphdr *iph, size_t len,
			    enum ip_statetype *statetype)
{
	struct protocol_info *info = (void *)state + protocol_offset;
	enum tcp_state newstate;
	struct tcphdr *tcph = (struct tcphdr *)((u_int32_t *)iph + iph->ihl);

	newstate = tcp_states[0][get_state_index(tcph)][TCP_STATE_NONE];

	/* Invalid: delete state */
	if (newstate == TCP_STATE_MAX) {
		DEBUGP("ip_state_tcp: invalid new deleting.\n");
		*statetype = IPS_INVALID;
		ip_state_delete(state);
		return NF_ACCEPT;
	}

	info->tcp_state = newstate;
	mod_timer(&state->timeout, jiffies + tcp_timeouts[info->tcp_state]);
	return NF_ACCEPT;
}

int ip_state_tcp_helper_register(struct ip_state_tcp_helper *helper)
{
	if (modreg_find(&helpers, (void *)(unsigned long)helper->port,
			&portcmp))
		return 0;

	modreg_prepend(&helpers, helper);
	MOD_INC_USE_COUNT;
	return 1;
}

void ip_state_tcp_helper_unregister(struct ip_state_tcp_helper *helper)
{
	IP_STATE_ASSERT(modreg_find(&helpers, 
				    (void *)(unsigned long)helper->port,
				    &portcmp)
			== helper);

	modreg_delete(&helpers, helper);
	MOD_DEC_USE_COUNT;
}

struct ip_state_protocol ip_state_protocol_tcp
= { NULL, IPPROTO_TCP, "tcp", sizeof(struct tcphdr), 
    tcp_pkt_to_tuple, tcp_invert_tuple, tcp_print_tuple, tcp_print_state, 
    tcp_destroy, tcp_established, tcp_new, 2 MINS};

__initfunc(void ip_state_protocol_tcp_init(void))
{
	protocol_offset = struct_reserve(ip_state, 
					 sizeof(struct protocol_info));
	DEBUGP("tcp: protocol_offset = %u\n", protocol_offset);
}
