/* All the functions to deal with the problem of creating a perconn
 * for a new packet. */
#include <stddef.h>

#include "bindings/ipnat_bind_generic.c"

/* Bind types; linked list kept in increasing private_offset order  */
static struct ip_nat_bind_type *bind_types;

typedef unsigned long NICE_TYPE;

static const struct ip_nat_binding *
find_binding(const struct ip_nat_pktinfo *pkt)
{
	const struct ip_nat_binding *ptr;

	for (ptr = conns.bind_table; ptr; ptr = ptr->next) {
		struct ip_nat_pktinfo mangled;
		unsigned int i;

		if (ptr->elem.info.hook != pkt->hook)
			continue;

		mangled = *pkt;
#if 1
		DEBUGP("pkt     bind    mask\n");    
#endif
		/* Apply mask. */
		for (i = 0; i < sizeof(ptr->elem.mask)/sizeof(NICE_TYPE); i++){
#if 1
			DEBUGP("%08lX %08lX %08lX\n",
			       ((NICE_TYPE *)pkt)[i], 
			       ((NICE_TYPE *)&ptr->elem.info)[i],
			       ((NICE_TYPE *)&ptr->elem.mask)[i]);
#endif
			((NICE_TYPE *)&mangled)[i]
				&= ((NICE_TYPE *)&ptr->elem.mask)[i];
		}

		if (memcmp(&ptr->elem.info, &mangled, sizeof(mangled)) == 0)
			break;
	}

	return ptr;
}

static unsigned int ip_nat_do_bind(const struct iphdr *iph,
				   const struct ip_nat_pktinfo *pkt,
				   struct ip_nat_perconn **perconn)
{
	const struct ip_nat_binding *bind = find_binding(pkt);

	if (!bind) {
		DEBUGP("ip_nat_do_bind: No binding for hook %s.\n",
		       hooknames[pkt->hook]);
		return NF_ACCEPT;
	} else {
		DEBUGP("ip_nat_do_bind: binding for hook %s.\n",
		       hooknames[pkt->hook]);
		return bind->elem.type[0]->bind(iph, pkt, &bind->elem,
						bind->mr, 0, perconn);
	}
}

static int bind_type_cmpfn(const void *arg, const void *curr)
{
	return strcmp(arg, ((struct ip_nat_bind_type *)curr)->name) == 0;
}

static struct ip_nat_bind_type *find_bind_type(const char *name, int load)
{
	struct ip_nat_bind_type *ret;

	ret = modreg_find(&bind_types, name, bind_type_cmpfn);
	if (!ret && load) {
		char modulename[IP_NAT_BIND_MAX_NAMELEN
			       + sizeof("ip_nat_bind_")];
		strcpy(modulename, "ip_nat_bind_");
		strcat(modulename, name);
		DEBUGP("ip_nat: requesting module `%s'\n", modulename);
		request_module(modulename);
		return find_bind_type(name, 0);
	}
	else return ret;
}

/* It'd really be easier to simply append to the end, knowing that
 * it'll all come out in the wash.  Of course, combined with module
 * unloading and autoloading, that'd be a *really cool* DoS.  If this
 * weren't free software, trust me, I'd have done it that way.  Maybe
 * there's a moral here... */
static void
find_private_space(struct ip_nat_bind_type *me)
{
	size_t last_offset = sizeof(struct ip_nat_perconn);
	struct ip_nat_bind_type **ptr;
 
	/* Go through, add type to wherever there's a big enough hole. */
	down(&modreg_sem);

	for (ptr = &bind_types; *ptr; ptr = &(*ptr)->next) {
		DEBUGP("find_private_space: %s used %u at offset %u\n",
		       (*ptr)->name, (*ptr)->private_size,
		       (*ptr)->private_offset);
		last_offset = (*ptr)->private_offset + (*ptr)->private_size;
		if ((*ptr)->next 
		    && ((*ptr)->next->private_offset - last_offset
			>= me->private_size)) {
			DEBUGP("find_private_space: found %u space\n",
			       me->private_size);
			me->private_offset = last_offset;
			me->next = (*ptr)->next;
			(*ptr)->next = me;
			up(&modreg_sem);
			return;
		}
	}

	/* We're at the end of the list. */
	me->private_offset = last_offset;
	__modreg_append(&bind_types, me);
	up(&modreg_sem);

	perconn_size = last_offset + me->private_size;
	DEBUGP("find_private_space: total size now %u.\n", perconn_size);
}

int ip_nat_register_bind_type(struct ip_nat_bind_type *me)
{
	if (find_bind_type(me->name, 0))
		return -EEXIST;

	MOD_INC_USE_COUNT;
	me->usage = 0;

	find_private_space(me);
	return 0;
}

/* Don't bother reducing perconn_size here; it'll be done on next
   find_private_space(). */
int ip_nat_unregister_bind_type(struct ip_nat_bind_type *me)
{
	if (me->usage) {
		DEBUGP("ip_nat_unregister_bind_type: Can't dereg %s\n",
		       me->name);
		return -EBUSY;
	}

	DEBUGP("ip_nat_unregister_bind_type: dereging %s\n", me->name);
	modreg_delete(&bind_types, me);
	MOD_DEC_USE_COUNT;
	DEBUGP("ip_nat_unregister_bind_type: dereg %s\n", me->name);
	return 0;
}

/* NAT bind tables are ordered, similar to routing tables: most
 * specific to least specific.  I've tried (and maybe failed) to be
 * intuitive here.  Higher the better.  This is a partial ordering.  */
static u_int32_t ip_nat_precedence(const struct ip_nat_pktinfo *mask)
{
	unsigned int i;
	u_int32_t prec = 0;

	/* We build this up most important to least. */

	/* Do we care about device?: 1 bit */
	prec |= mask->device ? 1 : 0;

	/* Source IP bits + destination IP bits: 7 bits */
	prec <<= 7;
	prec |= hweight32(mask->tuplepr.tuple.src.ip) 
		+ hweight32(mask->tuplepr.tuple.dst.ip);

	/* Do we care about protocol?: 1 bit */
	prec <<= 1;
	prec |= mask->tuplepr.proto ? 1 : 0;

	/* Per-protocol information: 1+5n bits */
	{
		u_int32_t perprotobits = 1;
		u_int32_t perprotoprec = 0;

		for (i = 0; 
		     i < sizeof(mask->tuplepr.tuple.src.u)/sizeof(u_int32_t); 
		     i++) {
			perprotoprec
				+= hweight32(mask->tuplepr.tuple.src.u.all[i]);
			perprotobits += 5;
		}

		for (i = 0; 
		     i < sizeof(mask->tuplepr.tuple.dst.u)/sizeof(u_int32_t); 
		     i++) {
			perprotoprec
				+= hweight32(mask->tuplepr.tuple.dst.u.all[i]);
			perprotobits += 5;
		}

		prec <<= perprotobits;
		prec |= perprotoprec;
	}

	return prec;
}

static int prec_cmpfn(const void *new, const void *curr)
{
	return ((struct ip_nat_binding *)new)->elem.precedence
		> ((struct ip_nat_binding *)curr)->elem.precedence;
}

/* Return 0 or -errno */
static int bind_user_to_bind(const struct ip_nat_bind_user *user,
			     struct ip_nat_binding **bind,
			     size_t size)
{
	unsigned int numtypes, i;
	const char *typename;
	struct ip_nat_multi_range *mr;

	if (user->hook != NF_IP_LOCAL_OUT 
	    && user->hook != NF_IP_POST_ROUTING
	    && user->hook != NF_IP_PRE_ROUTING) {
		DEBUGP("bind_user_to_bind: bad hook `%u'\n", user->hook);
		return -EINVAL;
	}

	if (~user->mask.hook) {
		DEBUGP("bind_user_to_bind: bad hook mask `%08X'\n", 
		       user->mask.hook);
		return -EINVAL;
	}

	/* Count the number of types. */
	for (typename = (char *)(user+1), numtypes = 0;
	     *typename; 
	     typename += strlen(typename)+1, numtypes++)
		DEBUGP("bind_user_to_bind: typename %i `%s'\n",
		       numtypes, typename);

	if (numtypes == 0) {
		DEBUGP("bind_user_to_bind: no types!\n");
		return -EINVAL;
	}

	if (!kmalloc_multi(GFP_USER, 
			   bind,
			   sizeof(struct ip_nat_binding)
			   + (numtypes+1)*sizeof(struct ip_nat_bind_type),
			   &mr,
			   sizeof(struct ip_nat_multi_range)
			   + sizeof(struct ip_nat_range),
			   (void *)NULL))
		return -ENOMEM;

	(*bind)->next = NULL;
	(*bind)->elem.mask = user->mask;
	(*bind)->elem.precedence = ip_nat_precedence(&user->mask);
	memcpy((*bind)->elem.ifname, user->ifname, 
	       sizeof((*bind)->elem.ifname));

	(*bind)->elem.info = ((struct ip_nat_pktinfo)
			      { dev_get(user->ifname), 
				user->hook,
				user->tupleproto });

	/* Set up range. */
	mr->rangesize = 1;
	mr->range[0] = user->range;
	(*bind)->mr = mr;

	/* Find types, and insert them. */
	for (typename = (char *)(user+1), i = 0;
	     *typename; 
	     typename += strlen(typename)+1, i++) {
		DEBUGP("Searching for type `%s'\n", typename);
		if (!((*bind)->elem.type[i] = find_bind_type(typename, 1))) {
			DEBUGP("bind_user_to_bind: can't find type `%s'\n", 
			       typename);
			kfree(*bind);
			return -ENOENT;
		}
	}
	(*bind)->elem.type[numtypes] = NULL;

	return 0;
}

/* FIXME: insert should add to range if bind already exists, and
   delete should delete from range. --RR */
static int insert_bind(struct ip_nat_bind_user *binduser, size_t len)
{
	int ret;
	struct ip_nat_binding *bind;

	/* double NUL terminate, in case it's bogus. */
	((char *)binduser)[len] = ((char *)binduser)[len+1] = '\0';
	
	ret = bind_user_to_bind(binduser, &bind, len);
	if (ret == 0) {
		unsigned int i;

		modreg_insert(&conns.bind_table, bind, prec_cmpfn);

		for (i = 0; bind->elem.type[i]; i++) {
			if (!bind->elem.type[i]->check(bind, i)) {
				DEBUGP("bind_user_to_bind: type `%s' upset\n", 
				       bind->elem.type[i]->name);
				kfree(bind);
				return -EINVAL;
			}
		}

		for (i = 0; bind->elem.type[i]; i++) {
			bind->elem.type[i]->usage++;
			if (bind->elem.type[i]->me)
				__MOD_INC_USE_COUNT(bind->elem.type[i]->me);
		}
	}
	return ret;
}

/* FIXME: Shouldn't compare the holes in the structure. --RR */
static int bind_cmpfn(const void *find, const void *curr)
{
	size_t offset = offsetof(struct ip_nat_binding, elem.info.tuplepr);

	/* Compare everything after device ptr in elem. */
	return memcmp(find + offset, curr + offset, 
		      sizeof(struct ip_nat_binding) - offset) == 0;
}

static void delete_bind(const struct ip_nat_binding *bind)
{
	unsigned int i;

	/* Blow up all the perconns created by this bind. */
	delete_perconns(&bind->elem);

	modreg_delete(&conns.bind_table, bind);

	for (i = 0; bind->elem.type[i]; i++) {
		bind->elem.type[i]->usage--;
		if (bind->elem.type[i]->me)
			__MOD_DEC_USE_COUNT(bind->elem.type[i]->me);
	}
	kfree(bind);
}

static int delete_bind_user(const struct ip_nat_bind_user *binduser,
			    size_t len)
{
	int ret;
	struct ip_nat_binding *bind;
	unsigned long flags;

	spin_lock_irqsave(&ip_nat_lock, flags);

	ret = bind_user_to_bind(binduser, &bind, len);
	if (ret == 0) {
		struct ip_nat_binding *match;
		match = modreg_find(&conns.bind_table, bind, bind_cmpfn);
		if (!match)
			ret = -ENOENT;
		else
			delete_bind(match);

		kfree(bind);
	}
	spin_unlock_irqrestore(&ip_nat_lock, flags);
	return ret;
}

static void ipnat_bind_down_device(const struct device *dev)
{
	struct ip_nat_binding *bind;
	unsigned long flags;

	spin_lock_irqsave(&ip_nat_lock, flags);
	for (bind = conns.bind_table; bind; bind = bind->next) {
		if (bind->elem.info.device == dev)
			bind->elem.info.device = NULL;
	}
	spin_unlock_irqrestore(&ip_nat_lock, flags);
}
	
static void ipnat_bind_up_device(const struct device *dev, const char *name)
{
	struct ip_nat_binding *bind;
	unsigned long flags;

	DEBUGP("Bringing up device %s (%p).\n", name, dev);
	spin_lock_irqsave(&ip_nat_lock, flags);
	for (bind = conns.bind_table; bind; bind = bind->next) {
		if (strcmp(bind->elem.ifname, name) == 0)
			bind->elem.info.device = dev;
	}
	spin_unlock_irqrestore(&ip_nat_lock, flags);
}

static int do_ip_nat_ctl(int cmd, void *user, unsigned int len)
{
	int ret;

	switch (cmd) {
	case IP_NAT_SO_DELETE:
	case IP_NAT_SO_INSERT: {
		struct ip_nat_bind_user *binduser;

		if (len < sizeof(struct ip_nat_binding)) {
			DEBUGP("IP_NAT_SO_INSERT/DELETE: len %u < %u\n", 
			       len, sizeof(struct ip_nat_binding));
			ret = -EINVAL;
		} else if (!(binduser = kmalloc(len+2, GFP_USER)))
			ret = -ENOMEM;
		else if (copy_from_user(binduser, user, len)) {
			kfree(binduser);
			ret = -EFAULT;
		} else {
			ret = (cmd == IP_NAT_SO_INSERT ? 
			       insert_bind(binduser, len)
			       : delete_bind_user(binduser, len));
			kfree(binduser);
		}
		break;
	}

	default:
		panic("ip_nat_bind: Someone wake Rusty.\n");
	}

	return ret;
}

#ifdef CONFIG_PROC_FS
static int dump_binding(char *buffer, const struct ip_nat_binding *bind)
{
	int len = 0;
	unsigned int i;

	len = sprintf(buffer + len, 
		      "%s %16s %u.%u.%u.%u/%lu -> %u.%u.%u.%u/%lu ",
		      hooknames[bind->elem.info.hook],
		      bind->elem.ifname, 
		      IP_PARTS(bind->elem.info.tuplepr.tuple.src.ip),
		      bind->elem.mask.tuplepr.tuple.src.ip == 0xFFFFFFFF ? 32 
		      : ffz(bind->elem.mask.tuplepr.tuple.src.ip),
		      IP_PARTS(bind->elem.info.tuplepr.tuple.dst.ip),
		      bind->elem.mask.tuplepr.tuple.dst.ip == 0xFFFFFFFF ? 32 
		      : ffz(bind->elem.mask.tuplepr.tuple.dst.ip));

	if (bind->elem.mask.tuplepr.proto)
		len += sprintf(buffer + len, "proto=%u ", 
			       bind->elem.info.tuplepr.proto);

	for (i = 0; bind->elem.type[i]; i++)
		len += sprintf(buffer + len, ":%s", bind->elem.type[i]->name);
	len += sprintf(buffer + len, "\n");

	return len;
}

/* List of perconns; offset in records not bytes. */
static int nat_get_binds(char *buffer, char **start, off_t offset,
			 int length, int dummy)
{
	struct ip_nat_binding *bind;
	off_t upto = 0;
	unsigned long flags;
	int len = 0,last_len = 0;

	spin_lock_irqsave(&ip_nat_lock, flags);
	for (bind = conns.bind_table; bind; bind = bind->next, upto++) {
		if (upto < offset) continue;

		last_len = len;
		len += dump_binding(buffer+len, bind);
		if (len > length) {
			len = last_len;
			break;
		}
	}

	spin_unlock_irqrestore(&ip_nat_lock, flags);
	buffer[len] = '\0';

	DEBUGP("nat_get_binds: Length = %i (of %i).  Offset = %li.\n",
	       len, length, upto);
	/* `start' hack - see fs/proc/generic.c line ~165 */
	*start = (char *)((unsigned int)upto - offset);
	return len;
}
#endif

static struct nf_setsockopt_ops ip_nat_sockopts 
= { NULL, PF_INET, IP_NAT_BASE_CTL, IP_NAT_SO_MAX, do_ip_nat_ctl };

static int init_ip_nat_bind(void)
{
	int ret;

	ret = ip_nat_register_bind_type(&bind_generic);
	if (ret == 0) {
		/* Generic doesn't count as usage. */
		MOD_DEC_USE_COUNT;
		ret = nf_register_sockopt(&ip_nat_sockopts);
		if (ret != 0)
			ip_nat_unregister_bind_type(&bind_generic);
	}
	return ret;
}

static void cleanup_ip_nat_bind(void)
{
	nf_unregister_sockopt(&ip_nat_sockopts);

	/* Delete all the binds. */
	while (conns.bind_table)
		modreg_delete(&conns.bind_table, conns.bind_table);

	/* Deregistering generic will dec. usage count. */ 
	MOD_INC_USE_COUNT;
	while (bind_types)
		ip_nat_unregister_bind_type(bind_types);
}

