/* Everything to do with perconn manipulation, particularly allocating
   them. */

/* The amount of total bytes required in a perconn. */
static size_t perconn_size = sizeof(struct ip_nat_perconn);

/* FIXME: It'd be nice to make hook a bitmask, so rules could apply to
   more than one chain (ie. LOCAL_OUT and PRE_ROUTING). --RR */
static struct ip_nat_manipulation *
find_manip(const struct ip_nat_tuple_proto *tuplepr, unsigned int hook)
{
	const struct ip_nat_manipulation *i;

	dump_tuplepr(__FUNCTION__, tuplepr, hook);

	for (i = *hash_by_tuple(tuplepr); i; i = i->next) {
		dump_tuplepr(__FUNCTION__ " looking at", 
			     &i->connection, i->hook);
		if (memcmp(tuplepr, &i->connection, sizeof(*tuplepr)) == 0
		    && hook == i->hook)
			break;
	}
	return (struct ip_nat_manipulation *)i;
}

void ip_nat_delete_perconn_timeout(struct ip_nat_perconn *perconn)
{
	/* Kill timer: interrupts are blocked, so it's no race. */
	if (perconn->timerfn) {
		DEBUGP("ip_nat_delete: del_timer %p.\n", &perconn->timeout);
		del_timer(&perconn->timeout);
	}

	/* Remove from the hashes */
	if (perconn->manip[IP_NAT_MANIP_SRC].perconn)
		modreg_delete(hash_by_tuple(&perconn->manip[IP_NAT_MANIP_SRC]
					    .connection),
			      &perconn->manip[IP_NAT_MANIP_SRC]);
	if (perconn->manip[IP_NAT_MANIP_DST].perconn)
		modreg_delete(hash_by_tuple(&perconn->manip
					    [IP_NAT_MANIP_DST].connection),
			      &perconn->manip[IP_NAT_MANIP_DST]);
	if (perconn->by_source.manip_info)
		modreg_delete(hash_by_source(&perconn->by_source.manip_info
					     ->manips[0].manip,
					     perconn->by_source.proto),
			      &perconn->by_source);
	modreg_delete(hash_by_ip_and_proto(perconn->by_ip_and_proto.srcip,
					   perconn->by_ip_and_proto.dstip,
					   perconn->by_ip_and_proto.proto),
		      &perconn->by_ip_and_proto);
	modreg_delete(hash_by_newtuple(&perconn->by_new_tuple[0].tuplepr,
				       perconn->by_new_tuple[0].hook),
		      &perconn->by_new_tuple[0]);
	modreg_delete(hash_by_newtuple(&perconn->by_new_tuple[1].tuplepr,
				       perconn->by_new_tuple[1].hook),
		      &perconn->by_new_tuple[1]);

	/* If it has a cause (ie. not null perconn), notify bindings
	 * of destruction */
	if (perconn->cause) {
		unsigned int i;

		for (i = 0; perconn->cause->type[i]; i++) 
			perconn->cause->type[i]->destroy(perconn);
	}

	kfree(perconn);
}

/* Never fails, returns NF_ACCEPT.  Suitable for use as ip_nat_fn. 
 * You need lock to use this. */
unsigned int ip_nat_delete_perconn(struct ip_nat_perconn *perconn,
				   struct iphdr *iph,
				   u_int16_t len,
				   enum ip_nat_direction dir)
{
	ip_nat_delete_perconn_timeout(perconn);
	return NF_ACCEPT;
}

unsigned int ip_nat_delete_maybe(struct ip_nat_perconn *perconn,
				 struct iphdr *iph,
				 u_int16_t len,
				 enum ip_nat_manip_type manip_type)
{
	/* Expire this direction.  If both expired, delete. */
	modreg_delete(hash_by_tuple(&perconn->manip[manip_type].connection),
		      &perconn->manip[manip_type]);
	perconn->manip[manip_type].perconn = NULL;

	if (!perconn->manip[!manip_type].perconn)
		ip_nat_delete_perconn_timeout(perconn);

	return NF_ACCEPT;
}

/* Null ip_nat_fn: returns NF_ACCEPT. */
unsigned int ip_nat_accept(struct ip_nat_perconn *perconn,
			   struct iphdr *iph,
			   u_int16_t len,
			   enum ip_nat_manip_type manip_type)
{
	return NF_ACCEPT;
}

/* ip_nat_fn: returns NF_ACCEPT, refreshes timers. */
unsigned int ip_nat_refresh(struct ip_nat_perconn *perconn,
			    struct iphdr *iph,
			    u_int16_t len,
			    enum ip_nat_manip_type manip_type)
{
	mod_timer(&perconn->timeout, jiffies + perconn->timeend);
	return NF_ACCEPT;
}

static void ipnat_timer_wrapper(unsigned long perconn_ul)
{
	struct ip_nat_perconn *perconn = (struct ip_nat_perconn *)perconn_ul;
	unsigned long flags;

	spin_lock_irqsave(&ip_nat_lock, flags);
	perconn->timerfn(perconn);
	spin_unlock_irqrestore(&ip_nat_lock, flags);
}

/* Returns NULL if can't alloc. */
struct ip_nat_perconn *
ip_nat_alloc_perconn(const struct ip_nat_pktinfo *pkt,
		     const struct ip_nat_binding_element *cause,
		     struct ip_nat_multi_range *mr,
		     ip_nat_fn *postmangle, ip_nat_fn *relatedpacket,
		     unsigned long timertime, 
		     void (*timerfn)(struct ip_nat_perconn *perconn))
{
	struct ip_nat_perconn *perconn;
	unsigned int manip_type = HOOK2MANIP(pkt->hook);
	struct ip_nat_tuple_proto new_tpr, inv_tpr;
	struct ip_nat_manip_info *minfo, *minfo_rev;

	/* I \/ GCC */
	static unsigned int other_hooks[] 
		= { [NF_IP_LOCAL_OUT] = NF_IP_LOCAL_IN,
		    [NF_IP_LOCAL_IN] = NF_IP_LOCAL_OUT,
		    [NF_IP_POST_ROUTING] = NF_IP_PRE_ROUTING,
		    [NF_IP_PRE_ROUTING] = NF_IP_POST_ROUTING
		};

	
	if (!(perconn = kmalloc(perconn_size, GFP_ATOMIC))) {
		DEBUGP("ip_nat_alloc_perconn: kmalloc failed\n");
		return NULL;
	}

	/* Initialize perconn */
	perconn->cause = cause;
	perconn->postmangle = postmangle;
	perconn->relatedpacket = relatedpacket;

	/* Fill in the new manips except for the packet information */
	perconn->manip[manip_type].next
		= perconn->manip[!manip_type].next
		= NULL;
	perconn->manip[manip_type].lasthit 
		= perconn->manip[!manip_type].lasthit
		= jiffies;
	perconn->manip[manip_type].hitcount
		= perconn->manip[!manip_type].hitcount
		= 0;
	perconn->manip[manip_type].perconn
		= perconn->manip[!manip_type].perconn
		= perconn;
	perconn->manip[manip_type].hook = pkt->hook;
	perconn->manip[!manip_type].hook = other_hooks[pkt->hook];

	if (!get_unique_tuplepr(&new_tpr, &pkt->tuplepr, mr, pkt->hook)) {
		DEBUGP("ip_nat_alloc_perconn: Can't get unique tuple.\n");
		kfree(perconn);
		return NULL;
	}

	/* We now have two tuples (SRCIP/SRCPT/DSTIP/DSTPT):
	   the original (A/B/C/D') and the mangled one (E/F/G/H').

	   We're only allowed to work with the SRC per-proto part,
	   so we create inverses of both to start, then derive the other
	   fields we need.
	*/

	/* Original connection: A/B/C/D' */
	perconn->manip[manip_type].connection = pkt->tuplepr;

	/* Reply connection: simply invert the new tuple (G/H/E/F') */
	perconn->manip[!manip_type].connection = new_tpr;
	invert_tuplepr(&perconn->manip[!manip_type].connection);

	/* Create inverse of original: C/D/A/B' */
	inv_tpr = pkt->tuplepr;
	invert_tuplepr(&inv_tpr);

#if 1
	dump_tuplepr("Original", &perconn->manip[manip_type].connection,
		     perconn->manip[manip_type].hook);

	dump_tuplepr("New", &new_tpr, perconn->manip[manip_type].hook);

	dump_tuplepr("Reply", &perconn->manip[!manip_type].connection,
		     perconn->manip[!manip_type].hook);
#endif

	/* Convenience variables. */
	minfo = &perconn->manip[manip_type].manip_info;
	minfo_rev = &perconn->manip[!manip_type].manip_info;

	minfo->num_manips = minfo_rev->num_manips = 0;
	/* E != A or F != B => a source manip. */
	if (memcmp(&new_tpr.tuple.src, &pkt->tuplepr.tuple.src,
		   sizeof(pkt->tuplepr.tuple.src)) != 0) {
		/* In this direction, a source manip. */
		minfo->manips[minfo->num_manips].manip_type
			= IP_NAT_MANIP_SRC;
		minfo->manips[minfo->num_manips].manip
			= new_tpr.tuple.src;
		minfo->num_manips++;

		/* In the reverse direction, a destination manip. */
		minfo_rev->manips[minfo_rev->num_manips].manip_type
			= IP_NAT_MANIP_DST;
		minfo_rev->manips[minfo_rev->num_manips].manip
			= pkt->tuplepr.tuple.src;
		minfo_rev->num_manips++;
	}

	/* G != C or H != D => a destination manip. */
	if (memcmp(&perconn->manip[!manip_type].connection.tuple.src,
		   &inv_tpr.tuple.src, sizeof(inv_tpr.tuple.src)) != 0) {
		/* In this direction, a destination manip */
		minfo->manips[minfo->num_manips].manip_type
			= IP_NAT_MANIP_DST;
		minfo->manips[minfo->num_manips].manip
			= perconn->manip[!manip_type].connection.tuple.src;
		minfo->num_manips++;

		/* In the reverse direction, a source manip. */
		minfo_rev->manips[minfo_rev->num_manips].manip_type
			= IP_NAT_MANIP_SRC;
		minfo_rev->manips[minfo_rev->num_manips].manip
			= inv_tpr.tuple.src;
		minfo_rev->num_manips++;
	}

	/* Initialize timer. */
	perconn->timerfn = timerfn;
	if (timerfn) {
		init_timer(&perconn->timeout);
		perconn->timeout.function = &ipnat_timer_wrapper;
		perconn->timeout.data = (unsigned long)perconn;
		perconn->timeout.expires = jiffies + timertime;
		add_timer(&perconn->timeout);
	}
	perconn->timeend = timertime;

	/* We store the SOURCE manipulation tuple info here, as an
           arbitrary choice. */
	perconn->by_ip_and_proto = ((struct ip_nat_by_ip_and_proto)
				    {NULL, 
				     perconn->manip[IP_NAT_MANIP_SRC]
				     .connection.proto,
				     perconn->manip[IP_NAT_MANIP_SRC]
				     .connection.tuple.src.ip,
				     perconn->manip[IP_NAT_MANIP_SRC]
				     .connection.tuple.dst.ip,
				     perconn
				    });

	/* Insert into the various hashes */
	modreg_prepend(hash_by_tuple(&perconn->manip[manip_type].connection),
		       &perconn->manip[manip_type]);

	modreg_prepend(hash_by_tuple(&perconn->manip[!manip_type].connection),
		       &perconn->manip[!manip_type]);

	modreg_prepend(hash_by_ip_and_proto(perconn->by_ip_and_proto.srcip,
					    perconn->by_ip_and_proto.dstip,
					    perconn->by_ip_and_proto.proto),
		       &perconn->by_ip_and_proto);

	/* This is the tuple after manipulation by this hook. */
	perconn->by_new_tuple[0] = ((struct ip_nat_by_new_tuple)
				    { NULL, pkt->hook, new_tpr });
	/* This is the tuple after manipulation by the reply hook. */
	perconn->by_new_tuple[1] = ((struct ip_nat_by_new_tuple)
				    { NULL, other_hooks[pkt->hook], inv_tpr });

	modreg_prepend(hash_by_newtuple(&perconn->by_new_tuple[0].tuplepr,
					perconn->by_new_tuple[0].hook),
		       &perconn->by_new_tuple[0]);
	modreg_prepend(hash_by_newtuple(&perconn->by_new_tuple[1].tuplepr,
					perconn->by_new_tuple[1].hook),
		       &perconn->by_new_tuple[1]);

	/* We store ip_nat_by_source iff IP_NAT_MANIP_SRC. */
	if (manip_type == IP_NAT_MANIP_SRC) {
		perconn->by_source = ((struct ip_nat_by_source)
				      { NULL, 
					perconn->manip[IP_NAT_MANIP_SRC]
					.connection.proto,
					perconn->manip[IP_NAT_MANIP_SRC]
					.connection.tuple.src,
					&perconn->manip[IP_NAT_MANIP_SRC]
					.manip_info });
		modreg_prepend(hash_by_source(&perconn->by_source.manip_info
					      ->manips[0].manip,
					      perconn->by_source.proto),
			       &perconn->by_source);
	}
	else perconn->by_source.manip_info = NULL;

	return perconn;
}

/* Delete all perconns created by a setup. */
static void delete_perconns(const struct ip_nat_binding_element *elem)
{
	unsigned int i;

	/* Traverse by_ip_and_proto hash; it has 1 entry per perconn,
           so it's easiest */
	for (i = 0; i < IP_NAT_HTABLE_SIZE; i++) {
		struct ip_nat_by_ip_and_proto *ipp;

		for (ipp = conns.by_ip_and_proto[i]; ipp; ) {
			struct ip_nat_by_ip_and_proto *next = ipp->next;

			if (ipp->perconn->cause == elem)
				ip_nat_delete_perconn_timeout(ipp->perconn);

			ipp = next;
		}
	}
}

#ifdef CONFIG_PROC_FS
static unsigned int print_single_manip(char *buffer,
				       const struct ip_nat_proto *proto,
				       const struct ip_nat_manip_info *minfo,
				       unsigned int num)
{
	unsigned int len;

	len = sprintf(buffer, "new %s=%u.%u.%u.%u ",
		      (minfo->manips[num].manip_type == IP_NAT_MANIP_SRC 
		       ? "src" : "dst"),
		      IP_PARTS(minfo->manips[num].manip.ip));

	len += proto->print_manip(buffer + len, &minfo->manips[num].manip.u);

	return len;
}

static unsigned int print_manip(char *buffer,
				const struct ip_nat_manipulation *manip)
{
	unsigned int len;
	struct ip_nat_proto *proto = find_proto(manip->connection.proto);
	unsigned long diff, sec, usec;

	diff = jiffies - manip->lasthit;
	usec = (diff % HZ) * 1000000 / HZ;
	sec = diff / HZ;

	len = sprintf(buffer, "%s: ", hooknames[manip->hook]);
	len += sprintf(buffer + len, "%slast=%lu.%06lu hits=%lu ", 
		       manip->perconn ? "" : "EXPIRED ",
		       sec, usec, manip->hitcount);
	len += sprintf(buffer + len, "src=%u.%u.%u.%u dst=%u.%u.%u.%u ", 
		       IP_PARTS(manip->connection.tuple.src.ip),
		       IP_PARTS(manip->connection.tuple.dst.ip));
	len += proto->print_tuple(buffer + len, &manip->connection.tuple);

	len += print_single_manip(buffer + len, proto, &manip->manip_info, 0);
	if (manip->manip_info.num_manips == 2) {
		len += sprintf(buffer + len, "EXTRA: ");
		len += print_single_manip(buffer + len, proto,
					  &manip->manip_info, 1);
	}

	return len;
}

static unsigned int print_perconn(char *buffer,
				  const struct ip_nat_perconn *perconn)
{
	unsigned int len = 0;
	struct ip_nat_proto *proto=find_proto(perconn->by_ip_and_proto.proto);

	if (perconn->cause) {
		unsigned int i;

		for (i = 0; perconn->cause->type[i]; i++) 
			len += sprintf(buffer + len, "%s:",
				       perconn->cause->type[i]->name);
	}
	else len += sprintf(buffer + len, "null:");
	len += sprintf(buffer + len, "%s:", proto->name);

	len += print_manip(buffer + len, &perconn->manip[IP_NAT_MANIP_SRC]);
	len += print_manip(buffer + len, &perconn->manip[IP_NAT_MANIP_DST]);

	if (perconn->cause) {
		unsigned int i;

		for (i = 0; perconn->cause->type[i]; i++) 
			len += perconn->cause->type[i]
				->print_private(buffer + len, perconn);
	}
	else len += proto->bindtype.print_private(buffer + len, perconn);

	len += sprintf(buffer + len, "\n");
	return len;
}

static int alloc_null_perconn(const struct iphdr *iph,
			      const struct ip_nat_pktinfo *pkt,
			      struct ip_nat_perconn **perconn)
{
	/* We want an exact range, but per-proto can vary iff neccesary. */
	struct {
		struct ip_nat_multi_range mr;
		struct ip_nat_range range;
	} range;

	range.mr.rangesize = 1;
	/* Don't care about per-proto part */
	range.range.flags = IP_NAT_RANGE_MAP_IPS;
	if (HOOK2MANIP(pkt->hook) == IP_NAT_MANIP_DST)
		range.range.min_ip
			= range.range.max_ip 
			= pkt->tuplepr.tuple.dst.ip;
	else
		range.range.min_ip
			= range.range.max_ip 
			= pkt->tuplepr.tuple.src.ip;

	return find_proto(iph->protocol)->bindtype.bind(iph, pkt,
							NULL, &range.mr,
							0, perconn);
}

static void cleanup_ip_nat_null_perconns(void)
{
	delete_perconns(NULL);
}

/* List of perconns; offset in records not bytes. */
static int nat_get_perconns(char *buffer, char **start, off_t offset,
			    int length, int dummy)
{
	unsigned int i;
	int len = 0,last_len = 0;
	off_t upto = 0;
	unsigned long flags;

	spin_lock_irqsave(&ip_nat_lock, flags);

	/* Traverse by_ip_and_proto hash; it has 1 entry per perconn,
           so it's easiest */
	for (i = 0; i < IP_NAT_HTABLE_SIZE; i++) {
		struct ip_nat_by_ip_and_proto *ipp;

		for (ipp = conns.by_ip_and_proto[i]; 
		     ipp; 
		     ipp = ipp->next, upto++) {
			if (upto < offset) continue;

			last_len = len;
			len += print_perconn(buffer + len, ipp->perconn);
			if (len > length) {
				len = last_len;
				goto finished;
			}
		}
	}

finished:
	spin_unlock_irqrestore(&ip_nat_lock, flags);
	buffer[len] = '\0';

	DEBUGP("nat_get_perconns: Length = %i (of %i).  Offset = %li.\n",
	       len, length, upto);
	/* `start' hack - see fs/proc/generic.c line ~165 */
	*start = (char *)((unsigned int)upto - offset);
	return len;
}
#endif

