#ifndef _IP_NAT_PERCONN_H
#define _IP_NAT_PERCONN_H
/* IP NAT per-connection stuff.  These rules are dynamically created
 * by the bind map, and kept in a hash table. */

/* Play with dynamic later; meanwhile if it's good enough for DaveM's
   TCP code... */
#define IP_NAT_HTABLE_SIZE 64
#include "ip_nat.h"
#include <linux/timer.h>

struct ip_nat_perconn;
struct ip_nat_binding_element;
struct ip_nat_ip_use;

typedef unsigned int ip_nat_fn(struct ip_nat_perconn *, 
			       struct iphdr *iph,
			       u_int16_t len,
			       enum ip_nat_manip_type);

/* Never fails, returns NF_ACCEPT.  Suitable for use as ip_nat_fn. 
 * You need lock to use this. */
extern unsigned int ip_nat_delete_perconn(struct ip_nat_perconn *perconn,
					  struct iphdr *iph,
					  u_int16_t len,
					  enum ip_nat_manip_type);

/* As above, but used for timer. */
extern void ip_nat_delete_perconn_timeout(struct ip_nat_perconn *perconn);

/* Expire in this direction.  If expired in both dirs, delete. */
extern unsigned int ip_nat_delete_maybe(struct ip_nat_perconn *perconn,
					struct iphdr *iph,
					u_int16_t len,
					enum ip_nat_manip_type);

/* Null ip_nat_fn: returns NF_ACCEPT. */
extern unsigned int ip_nat_accept(struct ip_nat_perconn *perconn,
				  struct iphdr *iph,
				  u_int16_t len,
				  enum ip_nat_manip_type);

/* ip_nat_fn: returns NF_ACCEPT, refreshes timer. */
extern unsigned int ip_nat_refresh(struct ip_nat_perconn *perconn,
				   struct iphdr *iph,
				   u_int16_t len,
				   enum ip_nat_manip_type);

/* Returns NULL if can't alloc. */
struct ip_nat_perconn *
ip_nat_alloc_perconn(const struct ip_nat_pktinfo *pkt,
		     const struct ip_nat_binding_element *cause,
		     struct ip_nat_multi_range *mr,
		     ip_nat_fn *postmangle, ip_nat_fn *relatedpacket,
		     unsigned long timertime, 
		     void (*timerfn)(struct ip_nat_perconn *perconn));

struct ip_nat_perconn;

struct ip_nat_manip_info
{
	/* Number of manipulations (usually 1, sometimes 0 (null
	   tuple) or 2 (local-out which changed route or its local-in
	   counterpart). */
	unsigned int num_manips;

	struct {
		/* The type of manipulation to perform */
		enum ip_nat_manip_type manip_type;

		/* Replacement. */
		struct ip_nat_tuple_manip manip;
	} manips[2];
};

struct ip_nat_manipulation
{
	/* Next pointer for tuple hash lists */
	struct ip_nat_manipulation *next;

	/* The tuple we match */
	struct ip_nat_tuple_proto connection;

	/* The hook we're on. */
	unsigned int hook;

	/* Creation or last-used time in jiffies. */
	unsigned long lasthit;

	/* Number of hits. */
	unsigned long hitcount;

	/* Perconn I belong to; NULL if no longer in hash. */
	struct ip_nat_perconn *perconn;

	struct ip_nat_manip_info manip_info;
};

/* To keep track of source NAT mappings */
struct ip_nat_by_source
{
	struct ip_nat_by_source *next;

	u_int32_t proto;
	struct ip_nat_tuple_manip src;

	struct ip_nat_manip_info *manip_info;
};

/* Used to figure out how many people using this IP & proto combination */
struct ip_nat_by_ip_and_proto
{
	struct ip_nat_by_ip_and_proto *next;

	u_int32_t proto;
	u_int32_t srcip;
	u_int32_t dstip;

	struct ip_nat_perconn *perconn;
};

/* Used to ensure mappings are unique. */
struct ip_nat_by_new_tuple
{
	struct ip_nat_by_new_tuple *next;

	unsigned int hook;
	struct ip_nat_tuple_proto tuplepr;
};

struct ip_nat_perconn
{
	/* We have two manipulations; indexed by HOOK2MANIP(hook). */
	struct ip_nat_manipulation manip[2];

	/* Whoever created us: may be NULL for nul perconns */
	const struct ip_nat_binding_element *cause;

	/* Most binds want a timer, so include it here. */
	unsigned long timeend;
	struct timer_list timeout;
	void (*timerfn)(struct ip_nat_perconn *perconn);

	/* Called when a packet matches: returns NF_*. */
	ip_nat_fn *postmangle;
	/* Called for related packets; return ignored. */
	ip_nat_fn *relatedpacket;

	struct ip_nat_by_ip_and_proto by_ip_and_proto;
	struct ip_nat_by_source by_source;
	struct ip_nat_by_new_tuple by_new_tuple[2];

	/* Private area hang off the end. */
};

/* Apply a manip to a tuple_proto. */
extern void ip_nat_apply_manip(struct ip_nat_tuple_proto *tuplepr,
			       const struct ip_nat_tuple_proto *orig,
			       const struct ip_nat_manip_info *manip_info);
#endif /*_IP_NAT_PERCONN_H*/
