/* NAT for ICMP.  This is not a separate module, as understanding of
 * ICMP is required for correct NAT functioning anyway.
 */
static int icmp_check(const struct ip_nat_binding *bind, unsigned int typenum)
{
	/* Must specify proto == ICMP. */
	if (bind->elem.mask.tuplepr.proto != 0xFF
	    || bind->elem.info.tuplepr.proto != IPPROTO_ICMP) {
		DEBUGP("icmp_bind_check: mask %u, proto = %u.\n",
		       bind->elem.mask.tuplepr.proto,
		       bind->elem.info.tuplepr.proto);
		return 0;
	}

	if (bind->elem.type[typenum+1]) {
		DEBUGP("icmp_bind_check: must be last.\n");
		return 0;
	}

	return 1;
}

/* FIXME: Put local in pktinfo --RR */
/* Function to call if we get a match */
static unsigned int icmp_bind(const struct iphdr *iph,
			      const struct ip_nat_pktinfo *pkt,
			      const struct ip_nat_binding_element *cause,
			      struct ip_nat_multi_range *mr,
			      unsigned int typenumber,
			      struct ip_nat_perconn **perconn)
{
	struct icmphdr *hdr = (struct icmphdr *)((u_int32_t *)iph + iph->ihl);

	DEBUGP("icmp_nat bind: Doing it to %u.\n", hdr->type);

	/* Can only deal with these types. */
	/* FIXME: If nul mapping, should just pass packet? --RR */
	if (hdr->type != ICMP_ECHO
	    && hdr->type != ICMP_TIMESTAMP
	    && hdr->type != ICMP_INFO_REQUEST
	    && hdr->type != ICMP_ADDRESS) {
		DEBUGP("icmp_nat bind: Can't NAT ICMP type %u.\n", hdr->type);
		return NF_DROP;
	}

	dump_tuplepr(__FUNCTION__, &pkt->tuplepr, pkt->hook);
	*perconn = ip_nat_alloc_perconn(pkt, cause, mr,
					&ip_nat_delete_maybe, &ip_nat_accept,
					HZ*60, &ip_nat_delete_perconn_timeout);
	if (!*perconn) {
		DEBUGP("icmp-nat: ip_nat_alloc_perconn returned NULL.\n");
		return NF_DROP;
	}
	else DEBUGP("icmp-nat: ip_nat_alloc_perconn returned ok.\n");
	return NF_ACCEPT;
}

static int icmp_pkt_to_tuple(const void *datah, 
			     size_t datalen, 
			     struct ip_nat_tuple *tuple)
{
	const struct icmphdr *hdr = datah;

	/* If we don't have the complete ICMP packet, no translate. */
	if ((char *)&hdr->un.echo.id + sizeof(hdr->un.echo.id) - (char *)datah
	    > datalen)
		return 0;
	tuple->dst.u.icmp.type = hdr->type;
	tuple->dst.u.icmp.code = hdr->code;
	tuple->src.u.icmp.id = hdr->un.echo.id;

	return 1;
}

static void icmp_manip_pkt(struct iphdr *iph, size_t len,
			   const struct ip_nat_tuple_manip *manip,
			   enum ip_nat_manip_type manip_type)
{
	struct icmphdr *hdr = (struct icmphdr *)((u_int32_t *)iph + iph->ihl);

	hdr->checksum = ip_nat_cheat_check(hdr->un.echo.id ^ 0xFFFF,
					   manip->u.icmp.id,
					   hdr->checksum);
	DEBUGP("Replacing icmp id %u with %u.\n", 
	       ntohs(hdr->un.echo.id), ntohs(manip->u.icmp.id));
	hdr->un.echo.id = manip->u.icmp.id;
}

static int icmp_in_range(const struct ip_nat_tuple_proto *tuplepr,
			 enum ip_nat_manip_type manip_type,
			 const union ip_nat_proto_manip *min,
			 const union ip_nat_proto_manip *max)
{
	return (tuplepr->tuple.src.u.icmp.id >= min->icmp.id
		&& tuplepr->tuple.src.u.icmp.id <= max->icmp.id);
}

static int icmp_unique_tuple(struct ip_nat_tuple_proto *tuplepr,
			     const struct ip_nat_range *range,
			     unsigned int hook)
{
	static u_int16_t id = 0;
	unsigned int range_size 
		= (unsigned int)range->max.icmp.id - range->min.icmp.id + 1;
	unsigned int i;

	for (i = 0; i < range_size; i++, id++) {
		tuplepr->tuple.src.u.icmp.id
			= range->min.icmp.id + (id % range_size);
		if (!ip_nat_used_tuple(tuplepr, hook)) {
			DEBUGP("icmp_unique_tuple: Allocating id %hu.\n", 
			       ntohs(tuplepr->tuple.src.u.icmp.id));
			return 1;
		}
		else DEBUGP("icmp_unique_tuple: id %hu already taken.\n", 
			    ntohs(tuplepr->tuple.src.u.icmp.id));
	}
	return 0;
}

/*
  0 Echo Reply <=> 8 Echo
  13 Timestamp <=> 14 Timestamp Reply
  15 Information Request <=> 16 Information Reply
  17 Address Mask Request <=> 18 Address Mask Reply */
static void icmp_invert_tuple(struct ip_nat_tuple *tuple)
{
	static u_int8_t invmap[] = { 8, 255, 255, 255, 255, 255, 255, 255,
				     0, 255, 255, 255, 255, 14, 13, 16,
				     15, 18, 17 };

	/* Should never set up perconn to handle non-invertable types */
	if (tuple->dst.u.icmp.type >= sizeof(invmap) 
	   || invmap[tuple->dst.u.icmp.type] == 255)
		panic("icmp_invert_tuple: can't handle type %u.\n",
		      tuple->dst.u.icmp.type);

	tuple->dst.u.icmp.type = invmap[tuple->dst.u.icmp.type];
}

static unsigned int 
icmp_print_tuple(char *buffer, const struct ip_nat_tuple *tuple)
{
	return sprintf(buffer, "type=%u code=%u id=%u ",
		       tuple->dst.u.icmp.type,
		       tuple->dst.u.icmp.code,
		       ntohs(tuple->src.u.icmp.id));
}

static unsigned int 
icmp_print_manip(char *buffer, const union ip_nat_proto_manip *prmanip)
{
	return sprintf(buffer, "id %u ", ntohs(prmanip->icmp.id));
}

static struct ip_nat_proto proto_icmp
= { NULL, IPPROTO_ICMP, "ICMP", 
    icmp_pkt_to_tuple, 
    icmp_manip_pkt, 
    icmp_in_range, 
    icmp_unique_tuple,
    icmp_invert_tuple, 
    icmp_print_tuple, 
    icmp_print_manip, 
    { NULL, "ICMP", 0, NULL, 0, 0, 
      NULL,
      &icmp_check,
      &icmp_bind,
      NULL }
};
