/* Program to manage IP Nat rules.
 *
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2 of the License, or
 *	(at your option) any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <errno.h>
#include <sys/types.h>
#include <net/if.h>
#include <netinet/in.h>
#include <getopt.h>
#include <string.h>
#include <stdlib.h>
#include <limits.h>
#include <stdarg.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <linux/netfilter_ipv4.h>
#include "ip_nat.h"
#include "ip_nat_rule.h"

static char version[] = "ipnatctl version 0.1";

#ifndef IPNAT_LIBDIR
#define IPNAT_LIBDIR "/usr/local/lib/ipnatctl"
#endif

#define NEW(type, contents...) 				\
({ 							\
	typeof (type *) _n = malloc(sizeof(type)); 	\
	if (_n) *_n = ( (type) ## contents);		\
	_n;						\
})

void print_usage(char *msg, ...) __attribute__((noreturn));

void print_usage(char *msg, ...)
{
	va_list args;

	va_start(args, msg);
	vfprintf(stderr, msg, args);
	va_end(args);

	fprintf(stderr,
		"\nUsage: "
		"   ipnatctl [-I|-D] <input spec> <output range> <bindings>\n"
		"   ipnatctl -L\n\n");

	fprintf(stderr,
		"<input spec>:\n"
		"  --source/-s address/mask		Source address\n"
		"  --destination/-d address/mask	Destination address\n"
		"  --interface/-i name			Interface to match\n"
		"                                       (dest manip->src iface)\n"
		"                                       (src manip->dest iface)\n"
		"  --local/-l				Local (excludes -i)\n"
		"  --manipulation-type/-m [source|destination]\n"
		"					Type of binding\n"
		"  --protocol/-p [name | number] [proto-specifics]\n"
		"					Protocol to match\n\n");

	fprintf(stderr,
		"<output range>:\n"
		"  --to/-t [address/mask | address-address]\n"
		"					IP address range to map to\n\n");

	fprintf(stderr,
		"<bindings>:\n"
		"  --bind <name>[,<name>...]		What to do with matches.\n");

	exit(1);
}

/* FIXME */
int list_rules(char *argv[])
{
	return system("cat /proc/net/ip_nat_rules");
}

/* From libops.c */
struct option *
try_load_library(const char *basename,
		 const char *lib, 
		 void (**help)(void), 
		 int (**parse)(int, char **, struct ip_nat_rule_user *),
		 struct option *original_opts);

static struct option original_opts[] = {
	{ "source", 1, 0, 's', },
	{ "destination", 1, 0, 'd' },
	{ "local", 1, 0, 'l' },
	{ "interface", 1, 0, 'i' },
	{ "protocol", 1, 0, 'p' },
	{ "manipulation-type", 1, 0, 'm' },
	{ "to", 1, 0, 't' },
	{ "binding", 1, 0, 'b' },
	{ "help", 0, 0, 'h' },
	{ "version", 0, 0, 'v' },
	{ NULL, 0, 0, 0 }
};

static int get_address(const char *string, u_int32_t *ip, u_int32_t *mask)
{
	struct in_addr addr;
	char *delim;

	delim = strchr(string, '/');
	if (delim) {
		struct in_addr addrmask;

		if (strchr(delim+1, '.')) {
			if (!inet_aton(delim+1, &addrmask))
				return 0;
			*mask = addrmask.s_addr;

			*delim = '\0';
			if (!inet_aton(string, &addr))
				return 0;
			*ip = addr.s_addr & *mask;

			return 1;
		}
		else {
			unsigned long tmpmask;

			errno = 0;
			tmpmask = strtoul(delim+1, NULL, 10);
			if (tmpmask == ULONG_MAX && errno)
				return 0;

			*mask = htonl((~0) << (32 - tmpmask));
			*delim = '\0';
			if (!inet_aton(string, &addr))
				return 0;

			*ip = addr.s_addr & *mask;

			return 1;
		}
	}
	else {
		if (!inet_aton(string, &addr))
			return 0;
		*ip = addr.s_addr;
	}

	*mask = ~0;
	return 1;
}

static void
get_source_address(struct ip_nat_rule_user *rule, const char *optarg)
{
	if (!get_address(optarg, &rule->tupleproto.tuple.src.ip,
			 &rule->mask.tuplepr.tuple.src.ip))
		print_usage("Bad source address `%s'", optarg);
}

static void
get_destination_address(struct ip_nat_rule_user *rule,
			const char *optarg)
{
	if (!get_address(optarg, &rule->tupleproto.tuple.dst.ip,
			 &rule->mask.tuplepr.tuple.dst.ip))
		print_usage("Bad destination address `%s'", optarg);
}

static void
get_interface(struct ip_nat_rule_user *rule, const char *optarg)
{
	if (strlen(optarg) > IFNAMSIZ-1)
		print_usage("Interface name `%s' too long", optarg);
	strcpy(rule->ifname, optarg);
	memset(&rule->mask.device, 0xFF, sizeof(rule->mask.device));
}

/* In case no /etc/protocols. */
struct pprot {
	char *name;
	unsigned short num;
};

static const struct pprot default_protos[] = {
	{ "all", 0 },
	{ "tcp", IPPROTO_TCP },
	{ "udp", IPPROTO_UDP },
	{ "icmp", IPPROTO_ICMP }
};

static char *
proto_to_name(u_int8_t proto)
{
	unsigned int i;

	if (proto) {
		struct protoent *pent = getprotobynumber(proto);
		if (pent)
			return pent->p_name;
	}

	for (i = 0; i < sizeof(default_protos)/sizeof(struct pprot); i++)
		if (default_protos[i].num == proto)
			return default_protos[i].name;

	return NULL;
}

static void
get_protocol(struct ip_nat_rule_user *rule, const char *optarg,
	     struct option **opts, 
	     void (**help)(void), 
	     int (**parse)(int, char **, struct ip_nat_rule_user *))
{
	int proto = atoi(optarg);

	if (proto < 1 || proto > 255) {
		struct protoent *pent;

		pent = getprotobyname(optarg);
		if (pent)
			proto = pent->p_proto;
		else {
			unsigned int i;

			for (i = 0; 
			     i < sizeof(default_protos)/sizeof(struct pprot);
			     i++) {
				if (strcasecmp(optarg, default_protos[i].name)
				    == 0)
					proto = default_protos[i].num;
			}

			if (i == sizeof(default_protos)/sizeof(struct pprot)) {
				print_usage("unknown protocol `%s' specified",
					    optarg);
			}
		}
	}

	rule->tupleproto.proto = proto;
	rule->mask.tuplepr.proto = 0xFF;

	if (proto_to_name(proto))
		*opts = try_load_library(IPNAT_LIBDIR "/libnatctl_proto_", 
					 proto_to_name(proto), help, parse,
					 *opts);
}

static void
get_manip_type(struct ip_nat_rule_user *rule, const char *optarg)
{
	rule->mask.hook = 0xFFFFFFFF;
	if (strncasecmp(optarg, "destination", strlen(optarg) ?: 1) == 0) {
		if (rule->hook != NF_IP_LOCAL_OUT)
			rule->hook = NF_IP_PRE_ROUTING;
	} else if (strncasecmp(optarg, "source", strlen(optarg) ?: 1) == 0) {
		if (rule->hook == NF_IP_LOCAL_OUT) 
			print_usage("Can't mix `-m source' with -l");
		rule->hook = NF_IP_POST_ROUTING;
	} 
	else 
		print_usage("Unknown manipulation type `%s'", optarg);
}

static void
set_local(struct ip_nat_rule_user *rule)
{
	if (rule->mask.device)
		print_usage("Can't use -l and -i.\n");
	if (rule->hook == NF_IP_POST_ROUTING)
		print_usage("Can't use -l and `-m source'");

	rule->hook = NF_IP_LOCAL_OUT;
	rule->mask.hook = 0xFFFFFFFF;
}

static int 
get_address_range(const char *string, u_int32_t *ipmin, u_int32_t *ipmax)
{
	struct in_addr minaddr;
	char *delim;

	delim = strchr(string, '-');
	if (delim) {
		struct in_addr maxaddr;

		if (!inet_aton(delim+1, &maxaddr))
			return 0;
		*ipmax = maxaddr.s_addr;
		*delim = '\0';
	}
	else if ((delim = strchr(string, '/')) != NULL) {
		if (strchr(delim+1, '.')) {
			struct in_addr addrmask;
			if (!inet_aton(delim+1, &addrmask))
				return 0;

			*delim = '\0';
			if (inet_aton(string, &minaddr) == 0)
				return 0;

			*ipmin = (minaddr.s_addr & addrmask.s_addr);
			*ipmax = (minaddr.s_addr | ~addrmask.s_addr);
		} else {
			unsigned long mask;
			errno = 0;
			mask = strtoul(delim+1, NULL, 10);
			if (mask == ULONG_MAX && errno)
				return 0;

			mask = htonl((~0) << (32 - mask));
			*delim = '\0';
			if (inet_aton(string, &minaddr) == 0)
				return 0;

			*ipmin = (minaddr.s_addr & mask);
			*ipmax = (minaddr.s_addr | ~mask);
			return (ntohl(*ipmax) >= ntohl(*ipmin));
		}
	}

	if (!inet_aton(string, &minaddr))
		return 0;
	*ipmin = minaddr.s_addr;
	if (!delim)
		*ipmax = *ipmin;

	return (ntohl(*ipmax) >= ntohl(*ipmin));
}

static void
get_to(struct ip_nat_rule_user *rule, const char *optarg)
{
	if (!get_address_range(optarg, 
			       &rule->range.min_ip, &rule->range.max_ip)) 
		print_usage("Bad address range `%s'", optarg);
	rule->range.flags |= IP_NAT_RANGE_MAP_IPS;
}

static void 
get_rules(struct ip_nat_rule_user **rule, 
	     char *optarg, 
	     size_t *size)
{
	char *name;
	unsigned int i = 0;

	*size = sizeof(**rule) + strlen(optarg) + 2;
	*rule = realloc(*rule, *size);

	name = strtok(optarg, ",");
	if (!name)
		print_usage("Expected a rule name");
		
	while (name) {
		strcpy((*rule)->typenames+i, name);
		i += strlen(name) + 1;

		name = strtok(NULL, ",");
	}
	/* Double-nul terminator. */
	(*rule)->typenames[i] = '\0';

	return;
}

enum
{
	SPECIFIED_HOOK = 1,
	SPECIFIED_PROTOCOL = 2
};

static struct ip_nat_rule_user *main_parse(int argc, char *argv[], size_t *size)
{
	struct ip_nat_rule_user *rule 
		= NEW(struct ip_nat_rule_user, { });
	struct option *opts = original_opts;
	int (*parse)(int, char **, struct ip_nat_rule_user *) = NULL;
	void (*help)(void) = NULL;
	int c;
	unsigned int option_flags = 0;
	const char *rules = NULL;

	*size = sizeof(struct ip_nat_rule_user);

	while ((c = getopt_long(argc, argv, "s:d:i:p:m:t:b:l", opts, NULL)) 
	       != -1) {
		switch (c) {
		case 's':
			get_source_address(rule, optarg);
			break;

		case 'd':
			get_destination_address(rule, optarg);
			break;

		case 'l':
			set_local(rule);
			option_flags |= SPECIFIED_HOOK;
			break;

		case 'i':
			get_interface(rule, optarg);
			break;

		case 'p':
			get_protocol(rule, optarg, &opts, &help, &parse);
			option_flags |= SPECIFIED_PROTOCOL;
			break;

		case 'm':
			get_manip_type(rule, optarg);
			option_flags |= SPECIFIED_HOOK;
			break;

		case 't':
			get_to(rule, optarg);
			break;

		case 'b':
			rules = optarg;
			break;

		case 'h':
			if (help) {
				help();
				exit(0);
			} else
				print_usage("ipnatctl help\n");

		case 'v':
			printf(version);
			exit(0);

		default:
			if (!parse || !parse(c, argv, rule))
				print_usage("Unknown argument `%s'", optarg);
		}
	}

	if (optind < argc)
		print_usage("unknown arguments found on commandline");

	/* FIXME: Don't use generic if they specified protocol, or
           tack "generic" on in the kernel --RR */
	{
		char rule_info[rules 
				 ? strlen(rules) + sizeof(",generic") + 1
				 : sizeof("generic") + 1];

		if (rules) {
			strcpy(rule_info, rules);
			strcat(rule_info, ",generic");
		} else {
			strcpy(rule_info, "generic");
		}
		get_rules(&rule, rule_info, size);
	}

	if (!(option_flags & SPECIFIED_HOOK))
		print_usage("Need --manipulation-type/-m option.");

	return rule;
}

int main(int argc, char *argv[])
{
	int sockfd, op;
	size_t size;
	static struct ip_nat_rule_user *rule;

	if (argc < 2)
		print_usage("Too few options.");

	if (strcmp(argv[1], "-I") == 0)
		op = IP_NAT_SO_INSERT;
	else if (strcmp(argv[1], "-D") == 0)
		op = IP_NAT_SO_DELETE;
	else if (strcmp(argv[1], "-L") == 0)
		return list_rules(argv+1);
	else print_usage("Unknown option `%s'", argv[1]);

	rule = main_parse(argc-1, argv+1, &size);

#if 0
	if (rule->range.flags & IP_NAT_RANGE_MAP_IPS)
	fprintf(stderr, "Range for --to: %u.%u.%u.%u - %u.%u.%u.%u.\n",
		ntohl(rule->range.min_ip) >> 24,
		(ntohl(rule->range.min_ip) >> 16) & 0xFF,
		(ntohl(rule->range.min_ip) >> 8) & 0xFF,
		ntohl(rule->range.min_ip) & 0xFF,
		ntohl(rule->range.max_ip) >> 24,
		(ntohl(rule->range.max_ip) >> 16) & 0xFF,
		(ntohl(rule->range.max_ip) >> 8) & 0xFF,
		ntohl(rule->range.max_ip) & 0xFF);

	if (rule->range.flags & IP_NAT_RANGE_PROTO_SPECIFIED)
		fprintf(stderr, "Ports %u to %u.\n",
			ntohs(rule->range.min.tcp.port),
			ntohs(rule->range.max.tcp.port));
#endif

	sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	if (sockfd < 0) {
		perror("Opening raw socket");
		exit(1);
	}

	if (setsockopt(sockfd, IPPROTO_IP, op, rule, size) < 0) {
		perror("setsockopt problem");
		return 1;
	}
	return 0;
}
