/* Kernel module to match TCP parameters. */
#include <linux/module.h>
#include <linux/tcp.h>
#include <linux/skbuff.h>

#include "ipt_tcp.h"
#include "ip_tables.h"

/* Returns 1 if the port is matched by the range, 0 otherwise */
static inline int 
port_match(u_int16_t min, u_int16_t max, u_int16_t port, int invert)
{
	return (port >= min && port <= max) ^ invert;
}

static int
find_option(u_int8_t option, 
	    const struct tcphdr *tcp, 
	    u_int16_t datalen,
	    int invert)
{
	unsigned int i = sizeof(struct tcphdr);
	const u_int8_t *opt = (u_int8_t *)tcp;

	/* If we don't have the whole header, can't be a match. */
	if (tcp->doff * 4 > datalen)
		return 0;

	while (i < tcp->doff * 4) {
		if (opt[i] == option) return !invert;
		if (opt[i] < 2) i++;
		else i += opt[1]?:1;
	}

	return invert;
}

static int 
match(struct sk_buff *skb, 
      const char *indev,
      const char *outdev,
      const union ipt_matchinfo *matchinfo,
      int offset,
      const void *hdr,
      u_int16_t datalen)
{
	const struct tcphdr *tcp = hdr;

	/* FIXME: Try tcp doff >> packet len against various stacks --RR */

#define FWINVTCP(bool,invflg) ((bool) ^ !!(matchinfo->tcp.invflags & invflg))

	/* Must be long enough to examine, and not a fragment. */
	return datalen >= sizeof(struct tcphdr)
		&& !offset 
		&& port_match(matchinfo->tcp.spts[0], matchinfo->tcp.spts[1],
			      ntohs(tcp->source),
			      !!(matchinfo->tcp.invflags & IPT_TCP_INV_SRCPT))
		&& port_match(matchinfo->tcp.dpts[0], matchinfo->tcp.dpts[1],
			      ntohs(tcp->dest),
			      !!(matchinfo->tcp.invflags & IPT_TCP_INV_DSTPT))
		&& FWINVTCP((((unsigned char *)tcp)[13] 
			     & matchinfo->tcp.flg_mask)
			    == matchinfo->tcp.flg_cmp, 
			    IPT_TCP_INV_FLAGS)
		&& (!matchinfo->tcp.option
		    || find_option(matchinfo->tcp.option, tcp, datalen,
				   matchinfo->tcp.invflags & IPT_TCP_INV_OPTION));
}

/* Called when user tries to insert an entry of this type. */
static int 
checkentry(const struct ipt_ip *ip,
	   const union ipt_matchinfo *matchinfo)
{
	/* Must specify proto == TCP, and no unknown invflags */
	return ip->proto == IPPROTO_TCP
		&& !(matchinfo->tcp.invflags & ~IPT_TCP_INV_MASK);
}

static struct ipt_match tcp_match
= { NULL, "tcp", &match, &checkentry, &__this_module };

int init_module(void)
{
	return ipt_register_match(&tcp_match);
}

void cleanup_module(void)
{
	ipt_unregister_match(&tcp_match);
}
