#include <linux/types.h>
#include <linux/module.h>
#include <linux/ip.h>
#include <linux/timer.h>
#include <linux/modreglist.h>
#include <net/protocol.h>
#include <net/checksum.h>
#include <linux/bitops.h>
#include <linux/proc_fs.h>
#include <asm/spinlock.h>
#include <linux/udp.h>
#include <linux/netfilter_ipv4.h>

#include "ip_nat_proto.h"
#include "ip_nat_rule.h"
#include "ip_nat_binding.h"

#define IP_NAT_UDP_ADD_USER (IP_NAT_SO_MAX+1)
#define IP_NAT_UDP_DELETE_USER (IP_NAT_SO_MAX+2)

#define IP_NAT_UDP_TIMEOUT (HZ*75)

struct ip_nat_udp_user_port
{
	struct ip_nat_udp_user_port *next;
	u_int16_t port;
	ip_nat_fn *matchfn;
	ip_nat_fn *relatedfn;
	void (*timerfn)(struct ip_nat_binding *binding);
	unsigned long timeout;
};

static struct ip_nat_udp_user_port *userports;

static int udp_check(const struct ip_nat_rule *rule, unsigned int typenum)
{
	/* Must specify proto == UDP. */
	if (rule->elem.mask.tuplepr.proto != 0xFF
	    || rule->elem.info.tuplepr.proto != IPPROTO_UDP) {
		DEBUGP("udp_rule_check: mask %u, proto = %u.\n",
		       rule->elem.mask.tuplepr.proto,
		       rule->elem.info.tuplepr.proto);
		return 0;
	}

	if (rule->elem.type[typenum+1]) {
		DEBUGP("udp_rule_check: must be last.\n");
		return 0;
	}

	return 1;
}

static struct ip_nat_udp_user_port no_user_port =
{ NULL, 0, &ip_nat_refresh, &ip_nat_delete_binding,
  &ip_nat_delete_binding_timeout, IP_NAT_UDP_TIMEOUT };

static int port_cmp(const void *arg, const void *curr)
{
	return ((struct ip_nat_udp_user_port *)curr)->port
		== (u_int16_t)(unsigned long)arg;
}

/* Function to call if we get a match */
static unsigned int udp_rule(const struct iphdr *iph,
			     const struct ip_nat_pktinfo *pkt,
			     const struct ip_nat_rule_element *cause,
			     struct ip_nat_multi_range *mr,
			     unsigned int typenumber,
			     struct ip_nat_binding **binding)
{
	struct udphdr *hdr = (struct udphdr *)((u_int32_t *)iph + iph->ihl);
	struct ip_nat_udp_user_port *user;

	/* Send to userspace if it's in the user port list. */
	user = modreg_find(&userports, (void *)(int)hdr->dest, &port_cmp);
	if (!user)
		user = &no_user_port;

	dump_tuplepr(__FUNCTION__, &pkt->tuplepr, pkt->hook);
	*binding = ip_nat_alloc_binding(pkt, cause, mr,
					user->matchfn, user->relatedfn,
					user->timeout,
					&ip_nat_delete_binding_timeout);

	if (!*binding) {
		DEBUGP("udp-nat: ip_nat_alloc_binding returned NULL.\n");
		return NF_DROP;
	}
	else DEBUGP("udp-nat: ip_nat_alloc_binding returned ok.\n");
	return NF_ACCEPT;
}

static int do_udp_ctl(int cmd, void *user, unsigned int len)
{
	struct ip_nat_udp_user_port *portinfo;

	switch (cmd) {
	case IP_NAT_UDP_ADD_USER:
		portinfo = kmalloc(sizeof(*portinfo), GFP_USER);
		if (!portinfo)
			return -ENOMEM;
		else {
			portinfo->port = (u_int16_t)(unsigned long)user;
			modreg_prepend(&userports, portinfo);
		}
		break;

	case IP_NAT_UDP_DELETE_USER:
		portinfo = modreg_find(&userports, user, &port_cmp);

		if (!portinfo)
			return -ENOENT;
		else {
			modreg_delete(&userports, portinfo);
			kfree(portinfo);
		}
	}

	return 0;
}

static int udp_in_range(const struct ip_nat_tuple_proto *tuplepr,
			enum ip_nat_manip_type manip_type,
			const union ip_nat_proto_manip *min,
			const union ip_nat_proto_manip *max)
{
	u_int16_t port;

	if (manip_type == IP_NAT_MANIP_SRC)
		port = tuplepr->tuple.src.u.udp.port;
	else
		port = tuplepr->tuple.dst.u.udp.port;

	return ntohs(port) >= ntohs(min->udp.port)
		&& ntohs(port) <= ntohs(max->udp.port);
}

static int udp_unique_tuple(struct ip_nat_tuple_proto *tuplepr,
			    const struct ip_nat_range *range,
			    unsigned int hook)
{
	static u_int16_t port = 0, *portptr;
	unsigned int range_size, min, i;

	if (HOOK2MANIP(hook) == IP_NAT_MANIP_SRC)
		portptr = &tuplepr->tuple.src.u.udp.port;
	else
		portptr = &tuplepr->tuple.dst.u.udp.port;

	/* If no range specified... */
	if (!(range->flags & IP_NAT_RANGE_PROTO_SPECIFIED)) {
		/* If it's dst rewrite, can't change port */
		if (HOOK2MANIP(hook) == IP_NAT_MANIP_DST)
			return 0;

		if (*portptr < 1024) {
			DEBUGP("Privileged Range map.\n");
			min = 1;
			range_size = 1023;
		} else {
			DEBUGP("Unprivileged Range map.\n");
			min = 1024;
			range_size = 65535 - 1024 + 1;
		}
	} else {
		min = ntohs(range->min.udp.port);
		range_size = ntohs(range->max.udp.port) - min + 1;
	}

	for (i = 0; i < range_size; i++, port++) {
		*portptr = htons(min + port % range_size);
		if (!ip_nat_used_tuple(tuplepr, hook)) {
			DEBUGP("Allocated port %u.\n", ntohs(*portptr));
			return 1;
		}
		else DEBUGP("port %u already taken.\n", ntohs(*portptr));
	}
	return 0;
}

static int udp_pkt_to_tuple(const void *datah, size_t datalen,
			    struct ip_nat_tuple *tuple)
{
	const struct udphdr *hdr = datah;
	if ((char *)&hdr->dest + sizeof(hdr->dest) - (char *)datah > datalen)
		return 0;
	tuple->src.u.udp.port = hdr->source;
	tuple->dst.u.udp.port = hdr->dest;

	return 1;
}

static void udp_manip_pkt(struct iphdr *iph, size_t len,
			  const struct ip_nat_tuple_manip *manip,
			  enum ip_nat_manip_type manip_type)
{
	struct udphdr *hdr = (struct udphdr *)((u_int32_t *)iph + iph->ihl);
	u_int32_t oldip, srcip, dstip;
	u_int16_t *portptr;

	if (manip_type == IP_NAT_MANIP_SRC) {
		/* Get rid of src ip and src pt */
		oldip = iph->saddr;
		portptr = &hdr->source;
		srcip = manip->ip;
		dstip = iph->daddr;
	} else {
		/* Get rid of dst ip and dst pt */
		oldip = iph->daddr;
		portptr = &hdr->dest;
		srcip = iph->saddr;
		dstip = manip->ip;
	}

	/* Don't redo checksums if it's not set. */
	if (hdr->check == 0) {
		*portptr = manip->u.udp.port;
		return;
	}

	/* Get rid of old src and dest ips, and old src and dest ports. */
	hdr->check = ip_nat_cheat_check(~oldip, manip->ip,
					ip_nat_cheat_check(*portptr ^ 0xFFFF,
							   manip->u.udp.port,
							   hdr->check));
	if (hdr->check == 0) hdr->check = 0xFFFF;
	*portptr = manip->u.udp.port;
}

static void udp_invert_tuple(struct ip_nat_tuple *tuple)
{
	/* Switch source and destination ports */
	u_int16_t tmp;

	tmp = tuple->src.u.udp.port;
	tuple->src.u.udp.port = tuple->dst.u.udp.port;
	tuple->dst.u.udp.port = tmp;
}

static unsigned int
udp_print_tuple(char *buffer, const struct ip_nat_tuple *tuple)
{
	return sprintf(buffer, "srcpt=%u dstpt=%u ",
		       ntohs(tuple->src.u.udp.port),
		       ntohs(tuple->dst.u.udp.port));
}

static unsigned int
udp_print_manip(char *buffer, const union ip_nat_proto_manip *prmanip)
{
	return sprintf(buffer, "port=%u ", ntohs(prmanip->udp.port));
}

/* FIXME: if it's handled by a per-proto, print that here. --RR */
#if 0
static unsigned int
udp_print_private(char *buffer, const struct ip_nat_binding *binding)
{
	return 0;
}
#endif

static struct ip_nat_proto proto_udp
= { NULL, IPPROTO_UDP, "UDP",
    udp_pkt_to_tuple,
    udp_manip_pkt,
    udp_in_range,
    udp_unique_tuple,
    udp_invert_tuple,
    udp_print_tuple,
    udp_print_manip,
    { NULL, "UDP", 0, &__this_module, 0, 0,
      NULL,
      &udp_check,
      &udp_rule,
      NULL }
};

static struct nf_setsockopt_ops ip_nat_udp_sockopts
= { NULL, PF_INET, IP_NAT_UDP_ADD_USER, IP_NAT_UDP_DELETE_USER, do_udp_ctl };

int init_module(void)
{
	int ret;
	DEBUGP("ip_nat_udp\n");

	ret = ip_nat_register_proto(&proto_udp);
	if (ret < 0) return ret;

	ret = nf_register_sockopt(&ip_nat_udp_sockopts);
	if (ret < 0)
		ip_nat_unregister_proto(&proto_udp);

	return ret;
}

void cleanup_module()
{
	ip_nat_unregister_proto(&proto_udp);
	nf_unregister_sockopt(&ip_nat_udp_sockopts);
}
