/* Kernel module to control the rate
 *
 * Jerome de Vivie   <devivie@info.enserb.u-bordeaux.fr>
 */
#include <linux/module.h>
#include <linux/skbuff.h>
#include "ip_tables.h"
#include "ipt_RATE.h"

#include <asm/spinlock.h>
#define IPT_RATE_SRC  0x01
#define IPT_RATE_DST  0x02

struct ipt_kern_rate {
	unsigned long rate, prev;
	unsigned short attack;
	u_int32_t src,dst;
	spinlock_t  lock;
};

#define MS2JIFFIES(ms) (((unsigned int)(ms)) * HZ / 1000)

static void
ipt_rate_log(const unsigned int flags, 
	     const unsigned int pos, const struct iphdr *iph)
{
	switch(flags) {
	case IPT_RATE_DST|IPT_RATE_SRC:
		printk("flood: %lu.%lu.%lu.%lu->%lu.%lu.%lu.%lu, rule %u\n",
		       (ntohl(iph->saddr)>>24)&0xFF,
		       (ntohl(iph->saddr)>>16)&0xFF,
		       (ntohl(iph->saddr)>>8)&0xFF,
		       (ntohl(iph->saddr))&0xFF,
		       (ntohl(iph->daddr)>>24)&0xFF,
		       (ntohl(iph->daddr)>>16)&0xFF,
		       (ntohl(iph->daddr)>>8)&0xFF,
		       (ntohl(iph->daddr))&0xFF,pos);
		break;
		
	case IPT_RATE_SRC:
		printk("flood: scan host from %lu.%lu.%lu.%lu, rule %u\n",
		       (ntohl(iph->saddr)>>24)&0xFF,
		       (ntohl(iph->saddr)>>16)&0xFF,
		       (ntohl(iph->saddr)>>8)&0xFF,
		       (ntohl(iph->saddr))&0xFF,pos);
		break;
		
	case IPT_RATE_DST:
		printk("flood: src spoofed, dst=%lu.%lu.%lu.%lu, rule %u\n",
		       (ntohl(iph->daddr)>>24)&0xFF,
		       (ntohl(iph->daddr)>>16)&0xFF,
		       (ntohl(iph->daddr)>>8)&0xFF,
		       (ntohl(iph->daddr))&0xFF,pos);
		break;
		
	default:
		printk("flood: random, rule %u\n",pos);
		break;
	}
	return;
}

static int 
ipt_rate_target(struct sk_buff *skb,
		const char *indev,
		const char *outdev,
		const union ipt_targinfo *info,
		unsigned int pos)
{
	struct ipt_rateinfo *r = (struct ipt_rateinfo *)info->other;
	struct ipt_kern_rate *kr = r->kr;
	struct iphdr *iph= skb->nh.iph;
	unsigned int flags = 0;
	unsigned long now = jiffies;
	
	spin_lock(&kr->lock);
	kr->rate += now - kr->prev;
	kr->prev = now;
	if (kr->rate > MS2JIFFIES(r->avg)) {
		kr->rate = MS2JIFFIES(r->avg);
		kr->attack = 0;
	}

	if (kr->rate >= r->peak) {
		kr->rate -= MS2JIFFIES(r->peak);
		spin_unlock(&kr->lock);
		return pos + 1;
	}

	if (r->log) {
		if (kr->attack > 1) {
			goto drop;
		} else if (!kr->attack) {
			kr->src = iph->saddr;
			kr->dst = iph->daddr;
		} else if (net_ratelimit()) {
			if(kr->src == iph->saddr)
				flags |= IPT_RATE_SRC;
			if(kr->dst == iph->daddr)
				flags |= IPT_RATE_DST;
			spin_unlock(&kr->lock);
			ipt_rate_log(flags,pos,iph);
			return -NF_DROP - 1;
		}
		kr->attack++;
	}

 drop:	
	spin_unlock(&kr->lock);
	return -NF_DROP - 1;
}

static int 
ipt_rate_checkentry(const union ipt_targinfo *info,
				 unsigned int pos,
				 unsigned int finalpos)
{
	struct ipt_rateinfo *r = (struct ipt_rateinfo *)info->other;
	struct ipt_kern_rate *kr;
	
	if(r->peak >= r->avg)
		return 0;
	
	kr = kmalloc(sizeof(struct ipt_kern_rate), GFP_USER);
	if(!kr)
		return 0;
  
	kr->prev = 0;
	kr->attack = 0;
	kr->rate = MS2JIFFIES(r->avg);
	kr->lock = SPIN_LOCK_UNLOCKED;
	r->kr = kr;

	return 1;
}

static void
ipt_rate_deleteentry(const union ipt_targinfo *info)
{
	struct ipt_rateinfo *r = (struct ipt_rateinfo *)info->other;
	kfree(r->kr);
}


static struct ipt_target ipt_rate_reg
= { NULL, "RATE", ipt_rate_target, ipt_rate_checkentry, ipt_rate_deleteentry, &__this_module };

int init_module(void)
{
	if (ipt_register_target(&ipt_rate_reg))
		return -EINVAL;
	return 0;
}

void cleanup_module(void)
{
	ipt_unregister_target(&ipt_rate_reg);
}
