/* Masquerade.  Simple setup which alters range to a local IP address
   (depending on route) before calling the next setup. */
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/timer.h>
#include <linux/module.h>
#include <linux/netfilter.h>
#include <net/protocol.h>
#include <net/checksum.h>
#include <linux/netfilter_ipv4.h>

#include "ip_nat_rule.h"

static unsigned int
masquerade_print(char *buffer, const struct ip_nat_binding *binding)
{
	return 0;
}

static int masquerade_check(const struct ip_nat_rule *rule,
			    unsigned int typenumber)
{
	/* We can't be the final one. */
	if (!rule->elem.type[typenumber+1]) {
		DEBUGP("ipnat_rule_masquerade_check: can't be final.\n");
		return 0;
	}

	if (HOOK2MANIP(rule->elem.info.hook) != IP_NAT_MANIP_SRC) {
		DEBUGP("ipnat_rule_masquerade_check: must be SRC manip.\n");
		return 0;
	}

	return 1;
}

static unsigned int masquerade_rule(const struct iphdr *iph,
				    const struct ip_nat_pktinfo *pkt,
				    const struct ip_nat_rule_element *cause,
				    struct ip_nat_multi_range *mr,
				    unsigned int typenumber,
				    struct ip_nat_binding **binding)
{
	unsigned int ret;
	unsigned int i;
	u_int32_t newdst;
	struct rtable *rt;
	struct ip_nat_multi_range *newrange;

	/* Pass 0 instead of saddr, since it's going to be changed
	   anyway. */
	/* FIXME: Maybe pass dev->ifindex (or dev->iflink?)  instead
	   of 0? --RR */
	if (ip_route_output(&rt, iph->daddr, 0, IPTOS_TOS(iph->tos), 0) != 0) {
		DEBUGP("ipnat_rule_masquerade: Can't reroute.\n");
		return NF_DROP;
	}
	newdst = inet_select_addr(rt->u.dst.dev, rt->rt_gateway,
				  RT_SCOPE_UNIVERSE);
	ip_rt_put(rt);

	newrange = kmalloc(sizeof(struct ip_nat_multi_range)
			   + sizeof(struct ip_nat_range)*mr->rangesize,
			   GFP_ATOMIC);
	if (!newrange) {
		DEBUGP("Masquerade: oom.\n");
		return NF_DROP;
	}

	/* Transfer from original range. */
	memcpy(newrange, mr,
	       sizeof(struct ip_nat_multi_range)
	       + sizeof(struct ip_nat_range) * mr->rangesize);

	/* Substitute IPs for new IP dst. */
	for (i = 0; i < mr->rangesize; i++) {
		newrange->range[i].flags |= IP_NAT_RANGE_MAP_IPS;
		newrange->range[i].max_ip = newrange->range[i].min_ip
			= newdst;
	}

	ret = cause->type[typenumber+1]->rule(iph, pkt, cause,
					      newrange, typenumber+1,
					      binding);

	kfree(newrange);
	return ret;
}

static struct ip_nat_mapping_type mapping_masquerade
= { NULL, "masquerade", 0, &__this_module, 0, 0,
    masquerade_print,
    masquerade_check,
    masquerade_rule,
    NULL };

int init_module(void)
{
	return ip_nat_register_mapping_type(&mapping_masquerade);
}

void cleanup_module(void)
{
	ip_nat_unregister_mapping_type(&mapping_masquerade);
}
