#include <linux/types.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/netfilter.h>
#include <linux/icmp.h>
#include "../ip_state_protocol.h"

#define ICMP_TIMEOUT (30*HZ)

static int icmp_pkt_to_tuple(const void *datah, size_t datalen,
			     struct ip_state_tuple *tuple)
{
	const struct icmphdr *hdr = datah;

	tuple->dst.u.icmp.type = hdr->type;
	tuple->src.u.icmp.id = hdr->un.echo.id;
	tuple->dst.u.icmp.code = hdr->code;

	return 1;
}

static int icmp_invert_tuple(struct ip_state_tuple *tuple,
			     const struct ip_state_tuple *orig)
{
	/* Add 1; spaces filled with 0. */
	static u_int8_t invmap[] 
		= { [ICMP_ECHO] = ICMP_ECHOREPLY + 1,
		    [ICMP_ECHOREPLY] = ICMP_ECHO + 1,
		    [ICMP_TIMESTAMP] = ICMP_TIMESTAMPREPLY + 1,
		    [ICMP_TIMESTAMPREPLY] = ICMP_TIMESTAMP + 1,
		    [ICMP_INFO_REQUEST] = ICMP_INFO_REPLY + 1,
		    [ICMP_INFO_REPLY] = ICMP_INFO_REQUEST + 1,
		    [ICMP_ADDRESS] = ICMP_ADDRESSREPLY + 1,
		    [ICMP_ADDRESSREPLY] = ICMP_ADDRESS + 1};

	if (orig->dst.u.icmp.type >= sizeof(invmap)
	    || !invmap[orig->dst.u.icmp.type])
		return 0;

	tuple->src.u.icmp.id = orig->src.u.icmp.id;
	tuple->dst.u.icmp.type = invmap[orig->dst.u.icmp.type] - 1;
	tuple->dst.u.icmp.code = orig->dst.u.icmp.code;
	return 1;
}

/* Print out the per-protocol part of the tuple. */
static unsigned int icmp_print_tuple(char *buffer, 
				     const struct ip_state_tuple *tuple)
{
	return sprintf(buffer, "type=%u code=%u id=%u ",
		       tuple->dst.u.icmp.type,
		       tuple->dst.u.icmp.code,
		       ntohs(tuple->src.u.icmp.id));
}

/* Print out the private part of the state. */
static unsigned int icmp_print_state(char *buffer, 
				     const struct ip_state *state)
{
	return 0;
}

static void icmp_destroy(struct ip_state *state)
{
}

/* Returns verdict for packet, and may modify statetype */
static unsigned int icmp_established(struct ip_state *state,
				     struct iphdr *iph, size_t len,
				     int direction, 
				     enum ip_statetype *statetype)
{
	ip_state_delete_maybe(state, direction);
	return NF_ACCEPT;
}

/* Called when a new connection for this protocol found. */
static unsigned int icmp_new(struct ip_state *state,
			     struct iphdr *iph, size_t len,
			     enum ip_statetype *statetype)
{
	static u_int8_t valid_new[]
		= { [ICMP_ECHO] = 1,
		    [ICMP_TIMESTAMP] = 1,
		    [ICMP_INFO_REQUEST] = 1,
		    [ICMP_ADDRESS] =  1 };

	if (state->tuplehash[0].tuple.dst.u.icmp.type >= sizeof(valid_new)
	    || !valid_new[state->tuplehash[0].tuple.dst.u.icmp.type]) {
		/* Can't create a new ICMP `conn' with this. */
		*statetype = IPS_INVALID;
		ip_state_delete(state);
	} else {
		/* We're a one-shot: delete this half. */
		ip_state_delete_maybe(state, 0);
	}
	return NF_ACCEPT;
}


static struct ip_state_protocol ip_state_icmp_protocol
= { NULL, IPPROTO_ICMP, "icmp", sizeof(struct icmphdr), 
    icmp_pkt_to_tuple, icmp_invert_tuple, icmp_print_tuple, icmp_print_state,
    icmp_destroy, icmp_established, icmp_new, ICMP_TIMEOUT };

__initfunc(void ip_state_icmp_protocol_init())
{
	ip_state_register_protocol(&ip_state_icmp_protocol);
}
