/* All about doing NAT for TCP. */
#include "ip_nat.h"
#include "ip_nat_bind.h"
#include "ip_nat_perconn.h"

#define IP_NAT_TCP_OPENING_TIMEOUT (75*HZ)
#define IP_NAT_TCP_ESTABLISHED_TIMEOUT (6*3600*HZ)
#define IP_NAT_TCP_CLOSED_TIMEOUT (75*HZ)

enum ip_nat_tcpstate {
	IP_NAT_TCP_OPENING,
	IP_NAT_TCP_ESTABLISHED,
	IP_NAT_TCP_CLOSING,
};

struct ip_nat_tcp_user_port
{
	struct ip_nat_tcp_user_port *next;
	u_int16_t port;
	ip_nat_fn *matchfn;
	ip_nat_fn *relatedfn;
	void (*timerfn)(struct ip_nat_perconn *perconn);
	unsigned long timeout;
};

static struct ip_nat_tcp_user_port *userports;

static struct ip_nat_tcp_user_port no_user_port =
{ NULL, 0, &process, &process_icmp, 
  IP_NAT_TCP_OPENING_TIMEOUT, &ip_nat_delete_perconn_timeout };

/* We tack this on the end of the perconn to track TCP state. */
struct ip_nat_perconn_tcp
{
	enum ip_nat_tcpstate state;
};

struct ip_nat_address_range_tcp
{
	struct ip_nat_address_range range;
	u_int16_t min_port, max_port;
};

/* Simple binding for TCP connections */
static unsigned int tcp_bind(const struct iphdr *iph,
			     const struct ip_nat_pktinfo *pkt,
			     enum ip_nat_direction direction,
			     const struct device *dev,
			     struct ip_nat_address_range *range,
			     const struct ip_nat_bind *bind,
			     unsigned int typenumber,
			     struct ip_nat_perconn **perconn,
			     size_t extrasize)
{
	struct tcphdr *hdr = (struct tcphdr *)((u_int32_t *)iph + iph->ihl);
	struct ip_nat_address_range_tcp range_tcp;
	struct ip_nat_tcp_user_port *portinfo;

	/* If it's a specific TCP rule, range will be a TCP range already,
	   otherwise create a sane one. */
	if (!bind->mask.tuplepr.proto) {
		range_tcp.range = 
			((struct ip_nat_address_range){ NULL, range->min_ip,
							range->max_ip, 0 });

		/* Keep low ports mapped to low ports, and high ports
		 * mapped to high ports. */
		if (ntohs(hdr->source) < 1024) {
			range_tcp.min_port = htons(1);
			range_tcp.max_port = htons(1023);
		} else {
			range_tcp.min_port = htons(1024);
			range_tcp.max_port = htons(65535);
		}

		range = &range_tcp.range;
	}

	portinfo = modreg_find(&userports, (void *)(unsigned long)hdr->dest, 
			       &port_cmp);
	if (!portinfo) portinfo = no_user_port;

	dump_tuplepr(__FUNCTION__, direction, &pkt->tuplepr);
	*perconn = ip_nat_alloc_perconn(cause, &pkt->tuplepr, direction,
					range.range,
					portinfo->matchfn,
					portinfo->relatedfn,
					portinfo->timeout,
					portinfo->timerfn,
					typenumber,
					extrasize
					+ sizeof(struct ip_nat_perconn_tcp));
	if (!*perconn) {
		DEBUGP("tcp-nat: ip_nat_alloc_perconn returned NULL.\n");
		return NF_DROP;
	}
	else DEBUGP("tcp-nat: ip_nat_alloc_perconn returned ok.\n");

	/* Set up my private region in perconn. */
	privates[typenumber] = (char *)(*perconn) + extrasize;
	*((struct ip_nat_perconn_tcp *)privates[typenumber])
		= ((struct ip_nat_perconn_tcp){ IP_NAT_TCP_OPENING });

	return NF_ACCEPT;
}

static unsigned int tcp_process_icmp(struct ip_nat_perconn *perconn,
				     void *private,
				     struct sk_buff *skb,
				     enum ip_nat_direction dir)
{
	struct ip_nat_perconn_tcp *perconn_tcp = (void *)(perconn + 1);
	struct icmphdr *hdr = (struct icmphdr *)((u_int32_t *)skb->nh.iph 
						 + skb->nh.iph->ihl);

	enum ip_nat_direction reply
		= (info->type == IP_NAT_TRANS_RNAT 
		   ? IP_NAT_OUTGOING : IP_NAT_INCOMING);

	if (perconn_tcp->state == IP_NAT_TCP_OPENING
	    && dir == reply
	    && hdr->type == ICMP_DEST_UNREACH) {
		mod_timer(&perconn->timeout, 
			  jiffies + IP_NAT_TCP_CLOSED_TIMEOUT);
		perconn_tcp->state = IP_NAT_TCP_CLOSING;
	}
	return NF_ACCEPT;
}

static unsigned int process(struct ip_nat_perconn *perconn,
			    struct sk_buff *skb,
			    enum ip_nat_direction dir)
{
	const struct iphdr *iph = skb->nh.iph;
	struct tcphdr *hdr = (struct tcphdr *)((u_int32_t *)iph + iph->ihl);
	struct ip_nat_perconn_tcp *perconn_tcp = (void *)(perconn + 1);
	struct ip_nat_tcp *info = (struct ip_nat_tcp *)perconn->cause->private;
	enum ip_nat_direction reply
		= (info->type == IP_NAT_TRANS_RNAT 
		   ? IP_NAT_OUTGOING : IP_NAT_INCOMING);

	switch (perconn_tcp->state) {
	case IP_NAT_TCP_OPENING:
		if (dir == reply) {
			/* First reply packet: RST -> move to CLOSING,
			 * Other -> move to ESTABLISHED. */
			if (!(ntohs(iph->frag_off) & IP_OFFSET)) {
				if (hdr->rst) {
					mod_timer(&perconn->timeout, 
						  jiffies
						  + IP_NAT_TCP_CLOSED_TIMEOUT);
					perconn_tcp->state 
						= IP_NAT_TCP_CLOSING;
				} else {
					mod_timer(&perconn->timeout, 
						  jiffies 
						  + IP_NAT_TCP_ESTABLISHED_TIMEOUT);
					perconn_tcp->state
						= IP_NAT_TCP_ESTABLISHED;
				}
			}
		} else {
			/* Refresh timer. */
			mod_timer(&perconn->timeout, 
				  jiffies + IP_NAT_TCP_OPENING_TIMEOUT);
		}
		break;

	case IP_NAT_TCP_CLOSING:
		/* Nothing can change it now; we're doomed. */
		return NF_ACCEPT;

	case IP_NAT_TCP_ESTABLISHED:
		/* If we see a RST or a FIN, enter expiry cycle. */
		if (!(ntohs(iph->frag_off) & IP_OFFSET) 
		    && (hdr->fin || hdr->rst)) {
			mod_timer(&perconn->timeout, 
				  jiffies + IP_NAT_TCP_CLOSED_TIMEOUT);
			perconn_tcp->state = IP_NAT_TCP_CLOSING;
		} else {
			/* Refresh timer */
			mod_timer(&perconn->timeout, 
				  jiffies + IP_NAT_TCP_ESTABLISHED_TIMEOUT);
		}
	}

	return NF_ACCEPT;
}

/* Hands packet to userspace. */
static unsigned int process_user(struct ip_nat_perconn *perconn,
				 struct sk_buff *skb,
				 enum ip_nat_direction dir)
{
	struct ip_nat_tcp *info = (struct ip_nat_tcp *)perconn->cause->private;
	struct tcphdr *hdr = (struct tcphdr *)((u_int32_t *)skb->nh.iph 
					       + skb->nh.iph->ihl);
	enum ip_nat_direction fromserver;

	process(perconn, skb, dir);

	if (info->type == IP_NAT_TRANS_MASQUERADE
	    || info->type == IP_NAT_TRANS_NAT)
		fromserver = IP_NAT_INCOMING;
	else
		fromserver = IP_NAT_OUTGOING;

	/* Set mark to dest port so they can have separate handlers. */ 
	if (dir == fromserver)
		skb->nfmark = hdr->source;
	else 
		skb->nfmark = hdr->dest;

	skb->nfreason = NF_SPECIAL_SET_BY_TCP_NAT;
	return NF_USERSPACE;
}

static int port_cmp(const void *arg, const void *curr)
{
	return ((struct ip_nat_tcp_user_port *)curr)->port 
		== (u_int16_t)(unsigned long)arg;
}

/* Function to call if we get a match */
unsigned int setup(const struct iphdr *iph, 
		   const struct ip_nat_pktinfo *pkt,
		   enum ip_nat_direction direction,
		   const struct ip_nat_setup *cause,
		   struct ip_nat_perconn **perconn,
		   int local)
{
	struct tcphdr *hdr = (struct tcphdr *)((u_int32_t *)iph + iph->ihl);
	struct ip_nat_tcp *info = (struct ip_nat_tcp *)cause->private;
	struct ip_nat_address_range_tcp range;
	struct ip_nat_perconn_tcp tcpinfo = { IP_NAT_TCP_OPENING };
	ip_nat_fn *matchfn;

	range.range.next = NULL;
	range.range.range_full = 0;
	/* Keep low ports mapped to low ports, and high ports mapped to
	 * high ports. */
	if (ntohs(hdr->source) < 1024) {
		range.min_port = htons(1);
		range.max_port = htons(1023);
	} else {
		range.min_port = htons(1024);
		range.max_port = htons(65535);
	}

	/* Send to userspace if it's in the user port list. */
	if (modreg_find(&userports, (void *)(unsigned long)hdr->dest, 
			&port_cmp))
		matchfn = &process_user;
	else matchfn = &process;

	/* Set range. */
	switch (info->type) {
	case IP_NAT_TRANS_MASQUERADE:
		if (!ip_nat_masq_set(iph, &range.range)) {
			DEBUGP("icmp-nat: unroutable packet.\n");
			dump_iph(__FUNCTION__, iph);
			return NF_DROP;
		}
		break;

	case IP_NAT_TRANS_REDIRECT:
		if (info->mapport) 
			range.min_port = range.max_port = info->mapport;
		else range.min_port = range.max_port = hdr->dest;
		if (!ip_nat_interface(pkt->device, &range.range))
			return NF_DROP;
		break;

	case IP_NAT_TRANS_RNAT:
		if (info->mapport) 
			range.min_port = range.max_port = info->mapport;
		else range.min_port = range.max_port = hdr->dest;
		/* Fall thru ... */
	case IP_NAT_TRANS_NAT:
		DEBUGP("tcp-nat: ip_min = %u.%u.%u.%u ip_max = %u.%u.%u.%u\n",
		       IP_PARTS(info->ipmin), 
		       IP_PARTS(info->ipmax));
		range.range.min_ip = info->ipmin;
		range.range.max_ip = info->ipmax;

		break;
	}

	dump_tuplepr(__FUNCTION__, direction, &pkt->tuplepr);
	*perconn = ip_nat_alloc_perconn(cause, &pkt->tuplepr, direction,
					local,
					&range.range,
					matchfn,
					&process_icmp,
					IP_NAT_TCP_OPENING_TIMEOUT,
					&ip_nat_delete_perconn_timeout,
					&tcpinfo, sizeof(tcpinfo));
	if (!*perconn) {
		DEBUGP("tcp-nat: ip_nat_alloc_perconn returned NULL.\n");
		return NF_DROP;
	}
	else DEBUGP("tcp-nat: ip_nat_alloc_perconn returned ok.\n");
	return NF_ACCEPT;
}

static int print_setup(char *buffer, const struct ip_nat_setup *setup)
{
	struct ip_nat_tcp *info = (struct ip_nat_tcp *)setup->private;
	int len;

	switch (info->type) {
	case IP_NAT_TRANS_MASQUERADE:
		len = sprintf(buffer, " MASQ");
		break;

	case IP_NAT_TRANS_REDIRECT:
		len = sprintf(buffer, " REDIR");
		if (info->mapport)
			len += sprintf(buffer + len, " to port %hu",
				       info->mapport);
		break;

	case IP_NAT_TRANS_NAT:
		len = sprintf(buffer, " NAT %u.%u.%u.%u",
			      IP_PARTS_NATIVE(info->ipmin));
		if (info->ipmin != info->ipmax)
			len += sprintf(buffer + len, "-%u.%u.%u.%u",
				       IP_PARTS_NATIVE(info->ipmax));
		break;

	default:
		len = sprintf(buffer," RNAT %u.%u.%u.%u", 
			      IP_PARTS_NATIVE(info->ipmin));
		if (info->ipmin != info->ipmax)
			len += sprintf(buffer + len, "-%u.%u.%u.%u",
				       IP_PARTS_NATIVE(info->ipmax));
		if (info->mapport)
			len += sprintf(buffer + len, " to port %hu",
				       info->mapport);
		break;

	}
	return len;
}

static int do_tcp_ctl(int cmd, void *user, unsigned int len)
{
	struct ip_nat_tcp_user_port *portinfo;

	switch (cmd) {
	case IP_NAT_TCP_ADD_USER:
		portinfo = kmalloc(sizeof(*portinfo), GFP_USER);
		if (!portinfo)
			return -ENOMEM;
		else {
			portinfo->port = (u_int16_t)(unsigned long)user;
			modreg_prepend(&userports, portinfo);
		}
		break;

	case IP_NAT_TCP_DELETE_USER:
		portinfo = modreg_find(&userports, user, &port_cmp);

		if (!portinfo)
			return -ENOENT;
		else {
			modreg_delete(&userports, portinfo);
			kfree(portinfo);
		}
	}

	return 0;
}

static struct nf_setsockopt_ops ip_nat_tcp_sockopts 
= { NULL, PF_INET, IP_NAT_TCP_ADD_USER, IP_NAT_TCP_DELETE_USER, do_tcp_ctl };

/* Print out the private part of the perconn (if any). */
static int print_perconn(char *buffer, const struct ip_nat_perconn *perconn)
{
	struct ip_nat_perconn_tcp *perconn_tcp = (void *)(perconn + 1);
	int len = 0;

	switch (perconn_tcp->state) {
	case IP_NAT_TCP_CLOSING:
		len = sprintf(buffer, " CLOSING");
		break;
	case IP_NAT_TCP_OPENING:
		len = sprintf(buffer, " OPENING");
		break;
	case IP_NAT_TCP_ESTABLISHED:
		break;
	}

	if (perconn->postmangle == &process_user)
		len += sprintf(buffer + len, " USER");

	return len;
}

static struct ip_nat_setup_type tcp_nat
= { NULL, "tcp-nat", 0, &__this_module, check, setup, print_setup, 
    print_perconn };

int init_module(void)
{
	int ret;
	DEBUGP("tcp-nat: setup = %p\n", setup);
	ret = ip_nat_register_setup_type(&tcp_nat);
	if (ret < 0) return ret;

	ret = nf_register_sockopt(&ip_nat_tcp_sockopts);
	if (ret < 0)
		ip_nat_unregister_setup_type(&tcp_nat);

	return ret;
}

void cleanup_module()
{
	ip_nat_unregister_setup_type(&tcp_nat);
	nf_unregister_sockopt(&ip_nat_tcp_sockopts);
}
