/* Helper library for per-protocol helpers. */
#include "libperproto.h"

/* Register to handle a port; returns -1 or netfilter_dev file handle.
 * If handle is -1, opens /dev/nf_ipv4, otherwise tries to attach that
 * protocol to the handle given.  */
int perproto_register(u_int8_t protocol, u_int16_t port, int direction, 
		      int handle)
{
	int newhandle;
	struct nfdev_head cond;
	u_int32_t hookmask, reason;

	/* Setup handle to receive packets first. */
	if (handle < 0) {
		cond = ((struct nfdev_head){ 0xFFFFFFFF, 0, 0 });
		newhandle = open("/dev/netfilter_ipv4", O_RDWR);

		/* Delete original condition. */
		if (newhandle > 0)
			if (ioctl(newhandle, NFDIOSDELCOND, &cond) < 0) {
				close(newhandle);
				return -1;
			}
		else return -1;
		/* FIXME: blow off any captured packets. */
	}
	else newhandle = handle;

	if (direction == IP_NAT_INCOMING) {
		hookmask = (1 << NF_IP_PRE_ROUTING)|(1 << NF_IP_LOCAL_OUT);
		if (protocol == IPPROTO_TCP)
			reason = NF_SPECIAL_SET_BY_TCP_NAT_IN;
		else 
			NF_SPECIAL_SET_BY_UDP_NAT_IN;
	} else {
		hookmask = (1 << NF_IP_POST_ROUTING);
		if (protocol == IPPROTO_TCP)
			reason = NF_SPECIAL_SET_BY_TCP_NAT_OUT;
		else 
			NF_SPECIAL_SET_BY_UDP_NAT_OUT;
	}

	cond = ((struct nfdev_head){ hookmask, port, reason });
	if (ioctl(newhandle, NFDIOSADDCOND, &cond) < 0) {
		if (handle == -1)
			close(newhandle);
		return -1;
	}
	return newhandle;
}
s
struct pp_pattern
{
	const unsigned char *pattern;
	size_t patternlen;

	size_t buffused;
	unsigned char buffer[0];
};

/* Register interest for a given data pattern. */
perproto_pattern_handle 
perproto_pattern_register(const void *pattern, size_t size)
{
	struct pp_pattern *pathandle
		= malloc(sizeof(*pathandle) + size);

	*pathandle = ((struct pp_pattern){ pattern, size, 0 });
	return pathandle;
}

/* Returns -1 or offset in packet. */
/* This implementation is naive.  Is it worth being smart? */
int
perproto_pattern_check(perproto_pattern_handle h, 
		       const unsigned char *data, u_int32_t len)
{
	unsigned int i;

	/* Seek match in buffer + data */
	for (i = 0; i < h->buffused; i++) {
		if (memcmp(h->buffer + i, h->pattern, h->buffused - i)
		    == 0) {
			if (len < h->patternlen - (h->buffused-i)) {
				/* Packet shorter than buffer. */
				if (memcmp(data, h->pattern + (h->buffused-i),
					   len) == 0) {
					/* Append to buffer. */
					memmove(h->buffer, h->buffer + i,
						h->buffused - i);
					h->buffused -= i;
					memcpy(h->buffer + h->buffused, data,
					       len);
					h->buffused += len;
					return -1;
				}
			} else {
				/* Packet longer than buffer. */
				if (memcmp(data, h->pattern + (h->buffused-i),
					   h->patternlen - (h->buffused-i))
				    == 0) {
					i = h->patternlen - (h->buffused-i);
					h->buffused = 0;
					return i;
				}
			}
		}
	}

	/* No partial. */
	if (len > h->patternlen) {
		/* Do we match in this packet? */
		for (i = 0; i < len - h->patternlen; i++) {
			if (memcmp(data+i, h->pattern, h->patternlen) == 0) {
				h->buffused = 0;
				return i + h->patternlen;
			}
		}
	}

	/* Look for end match. */
	if (len < h->patternlen)
		i = 0;
	else 
		i = len - h->patternlen;
	for (; i < len; i++) {
		if (memcmp(data + i, h->pattern, len - i) == 0) {
			memcpy(h->buffer, data+i, len - i);
			h->buffused = len - i;
			return -1;
		}
	}
	h->buffused = 0;
	return -1;
}

/* Unregister interest in a given pattern */
void
perproto_pattern_unregister(perproto_pattern_handle handle)
{
	free(handle);
}

/* Given this protocol, srcip, dstip, srcpt and dstpt (host order),
   and direction, return the replacement ip/port.  Returns bool. */
int perproto_getorig(u_int8_t proto, u_int32_t srcip, u_int32_t dstip, 
		     u_int16_t srcpt, u_int16_t dstpt, int direction,
		     u_int32_t *ip, u_int16_t *pt)
{
	/* Look through /proc for connection entry. */
	FILE *fp = fopen("/proc/net/ip_nat_connections", "r");
	char line[200];

	if (!fp)
		return 0;

	/* We read these in as ascii, so they'll be in host order. */
	srcpt = ntohs(srcpt);
	dstpt = ntohs(dstpt);
	
	while (fgets(line, sizeof(line), fp)) {
		char *sip, *dip, *oldip, *spt, *dpt, *oldpt;

		if (proto == IPPROTO_TCP && strcmp(line, "tcp_nat ") != 0)
			continue;
		else if (proto == IPPROTO_UDP && strcmp(line, "udp_nat ") != 0)
			continue;

		if (strcmp(line + strlen(line) - 6, " USER\n") != 0)
			continue;

		if (direction == IP_NAT_INCOMING) {
			char *in_ptr = strstr(line, " in:") + 4;
			/* Old IP is source address. */
			oldip = strstr(in_ptr, " src=") + 5;
			sip = strstr(in_ptr, " new ") + 9;
			dip = strstr(in_ptr, " dst=") + 5;

			oldpt = strstr(in_ptr, " srcpt=") + 7;
			spt = strstr(in_ptr, " port=") + 6;
			dpt = strstr(in_ptr, " dstpt=") + 7;
		} else {
			char *in_ptr = strstr(line, " out:") + 5;
			/* Old IP is dest address. */
			oldip = strstr(in_ptr, " dst=") + 5;
			sip = strstr(in_ptr, " src=") + 5;
			dip = strstr(in_ptr, " new ") + 9;
			
			oldpt = strstr(in_ptr, " dstpt=") + 7;
			spt = strstr(in_ptr, " srcpt=") + 7;
			dpt = strstr(in_ptr, " port=") + 6;
		}

		if (inet_addr(sip) == srcip
		    && inet_addr(dip) == dstip
		    && atoi(dpt) == dstpt
		    && atoi(spt) == srcpt) {
			*ip = inet_addr(oldip);
			*pt = htons(atoi(oldpt));
			fclose(fp);
			return 1;
		}
	}

	fclose(fp);
	return 0;
}
