#include <linux/types.h>
#include <linux/module.h>
#include <linux/ip.h>
#include <linux/timer.h>
#include <linux/modreglist.h>
#include <net/protocol.h>
#include <net/checksum.h>
#include <linux/bitops.h>
#include <linux/proc_fs.h>
#include <asm/spinlock.h>
#include <linux/netfilter.h>
#include <linux/udp.h>

#include "ip_nat_proto.h"
#include "ip_nat_bind.h"
#include "ip_nat_perconn.h"

#define IP_NAT_UDP_ADD_USER (IP_NAT_SO_MAX+1)
#define IP_NAT_UDP_DELETE_USER (IP_NAT_SO_MAX+2)

#define IP_NAT_UDP_TIMEOUT (HZ*75)

/* FIXME: Do user lookup. --RR */
struct ip_nat_udp_user_port
{
	struct ip_nat_udp_user_port *next;
	u_int16_t port;
	ip_nat_fn *matchfn;
	ip_nat_fn *relatedfn;
	void (*timerfn)(struct ip_nat_perconn *perconn);
	unsigned long timeout;
};

static struct ip_nat_udp_user_port *userports;

static int udp_check(const struct ip_nat_binding *bind, unsigned int typenum)
{
	/* Must specify proto == UDP. */
	if (bind->elem.mask.tuplepr.proto != 0xFF
	    || bind->elem.info.tuplepr.proto != IPPROTO_UDP) {
		DEBUGP("udp_bind_check: mask %u, proto = %u.\n",
		       bind->elem.mask.tuplepr.proto,
		       bind->elem.info.tuplepr.proto);
		return 0;
	}

	if (bind->elem.type[typenum+1]) {
		DEBUGP("udp_bind_check: must be last.\n");
		return 0;
	}

	return 1;
}

/* We don't care when you destroy perconns. */
static void udp_destroy(const struct ip_nat_perconn *perconn)
{
}

static struct ip_nat_udp_user_port no_user_port =
{ NULL, 0, &ip_nat_refresh, &ip_nat_delete_perconn, 
  &ip_nat_delete_perconn_timeout, IP_NAT_UDP_TIMEOUT };

static int port_cmp(const void *arg, const void *curr)
{
	return ((struct ip_nat_udp_user_port *)curr)->port 
		== (u_int16_t)(unsigned long)arg;
}

/* Function to call if we get a match */
static unsigned int udp_bind(const struct iphdr *iph,
			     const struct ip_nat_pktinfo *pkt,
			     const struct ip_nat_binding_element *cause,
			     struct ip_nat_multi_range *range,
			     unsigned int typenumber,
			     struct ip_nat_perconn **perconn)
{
	struct udphdr *hdr = (struct udphdr *)((u_int32_t *)iph + iph->ihl);
	struct ip_nat_udp_user_port *user;

	/* Send to userspace if it's in the user port list. */
	user = modreg_find(&userports, (void *)(int)hdr->dest, &port_cmp);
	if (!user)
		user = &no_user_port;

	dump_tuplepr(__FUNCTION__, pkt->dir, &pkt->tuplepr);
	*perconn = ip_nat_alloc_perconn(pkt, cause, range,
					user->matchfn, user->relatedfn,
					user->timeout, 
					&ip_nat_delete_perconn_timeout);
	if (!*perconn) {
		DEBUGP("udp-nat: ip_nat_alloc_perconn returned NULL.\n");
		return NF_DROP;
	}
	else DEBUGP("udp-nat: ip_nat_alloc_perconn returned ok.\n");
	return NF_ACCEPT;
}

static int do_udp_ctl(int cmd, void *user, unsigned int len)
{
	struct ip_nat_udp_user_port *portinfo;

	switch (cmd) {
	case IP_NAT_UDP_ADD_USER:
		portinfo = kmalloc(sizeof(*portinfo), GFP_USER);
		if (!portinfo)
			return -ENOMEM;
		else {
			portinfo->port = (u_int16_t)(unsigned long)user;
			modreg_prepend(&userports, portinfo);
		}
		break;

	case IP_NAT_UDP_DELETE_USER:
		portinfo = modreg_find(&userports, user, &port_cmp);

		if (!portinfo)
			return -ENOENT;
		else {
			modreg_delete(&userports, portinfo);
			kfree(portinfo);
		}
	}

	return 0;
}

/* Because UDP, UDP and ICMP fields of struct ip_nat_address_range (hangs
   off end) and ip_nat_manip_tuple are structurally equivalent, these
   next two functions are identical for all three, so we cheat and use
   one for all */
static int udp_in_range(const struct ip_nat_tuple *tuple, 
			const struct ip_nat_range *range)
{
	return !range->proto_specific 
		|| (ntohs(tuple->src.u.udp.port) >= ntohs(range->min.udp.port) 
		    && ntohs(tuple->src.u.udp.port)
		    <= ntohs(range->max.udp.port));
}

static int udp_unique_tuple(struct ip_nat_tuple_proto *tuplepr, 
			    const struct ip_nat_range *range)
{
	static u_int16_t port = 0;
	unsigned int range_size;
	struct ip_nat_range newrange;
	unsigned int i;

	/* Try no change, if in range. */
	if (udp_in_range(&tuplepr->tuple, range)
	    && !ip_nat_used_tuple(tuplepr)) {
		DEBUGP("Straight thru port %hu.\n",
		       ntohs(tuplepr->tuple.src.u.udp.port));
		return 1;
	}

	/* If no range specified:
	    If it's < 1024, keep it there.
	    Otherwise keep it >= 1024. */
	if (!range->proto_specific) {
		newrange = *range;
		newrange.proto_specific = 1;
		if (tuplepr->tuple.src.u.udp.port < 1024) {
			DEBUGP("Privileged Range map.\n");
			newrange.min.udp.port = htons(0);
			newrange.max.udp.port = htons(1023);
		} else {
			DEBUGP("Unprivileged Range map.\n");
			newrange.min.udp.port = htons(1024);
			newrange.max.udp.port = htons(65535);
		}
		range = &newrange;
	}

	range_size = (unsigned int)ntohs(range->max.udp.port)
		- ntohs(range->min.udp.port) + 1;

	for (i = 0; i < range_size; i++, port++) {
		tuplepr->tuple.src.u.udp.port
			= htons(ntohs(range->min.udp.port) 
				+ (port % range_size));
		if (!ip_nat_used_tuple(tuplepr)) {
			DEBUGP("Allocated port %u.\n", 
			       ntohs(tuplepr->tuple.src.u.udp.port));
			return 1;
		}
		else DEBUGP("port %u already taken.\n", 
			    ntohs(tuplepr->tuple.src.u.udp.port));
	}
	return 0;
}

static int udp_pkt_to_tuple(const void *datah, size_t datalen, 
			    struct ip_nat_tuple *tuple)
{
	const struct udphdr *hdr = datah;
	if ((char *)&hdr->dest + sizeof(hdr->dest) - (char *)datah > datalen)
		return 0;
	tuple->src.u.udp.port = hdr->source;
	tuple->dst.u.udp.port = hdr->dest;

	return 1;
}

static void udp_manip_pkt(struct iphdr *iph, size_t len,
			  const struct ip_nat_tuple_manip *manip,
			  enum ip_nat_direction dir)
{
	struct udphdr *hdr = (struct udphdr *)((u_int32_t *)iph + iph->ihl);
	u_int32_t oldip, srcip, dstip;
	u_int16_t *portptr;

	if (dir == IP_NAT_OUTGOING) {
		/* Get rid of src ip and src pt */
		oldip = iph->saddr;
		portptr = &hdr->source;
		srcip = manip->ip;
		dstip = iph->daddr;
	} else {
		/* Get rid of dst ip and dst pt */
		oldip = iph->daddr;
		portptr = &hdr->dest;
		srcip = iph->saddr;
		dstip = manip->ip;
	}

	/* Don't redo checksums if it's not set. */
	if (hdr->check == 0) {
		*portptr = manip->u.udp.port;
		return;
	}

	/* Get rid of old src and dest ips, and old src and dest ports. */
	hdr->check = ip_nat_cheat_check(~oldip, manip->ip,
					ip_nat_cheat_check(*portptr ^ 0xFFFF, 
							   manip->u.udp.port,
							   hdr->check));
	if (hdr->check == 0) hdr->check = 0xFFFF;
	*portptr = manip->u.udp.port;
}

static void udp_invert_tuple(struct ip_nat_tuple *tuple)
{
	/* Switch source and destination ports */
	u_int16_t tmp;

	tmp = tuple->src.u.udp.port;
	tuple->src.u.udp.port = tuple->dst.u.udp.port;
	tuple->dst.u.udp.port = tmp;
}

static unsigned int 
udp_print_tuple(char *buffer, const struct ip_nat_tuple *tuple)
{
	return sprintf(buffer, " srcpt=%u dstpt=%u",
		       ntohs(tuple->src.u.udp.port), 
		       ntohs(tuple->dst.u.udp.port));
}

static unsigned int
udp_print_manip(char *buffer, const struct ip_nat_tuple_manip *manip)
{
	return sprintf(buffer, " port=%u", ntohs(manip->u.udp.port));
}

static unsigned int 
udp_print_private(char *buffer, const struct ip_nat_perconn *perconn)
{
	/* FIXME: if it's handled by a per-proto, print that here. --RR */
	return 0;
}

static struct ip_nat_proto proto_udp
= { NULL, IPPROTO_UDP, "UDP", 
    udp_pkt_to_tuple, 
    udp_manip_pkt, 
    udp_in_range, 
    udp_unique_tuple,
    udp_invert_tuple, 
    udp_print_tuple, 
    udp_print_manip, 
    { NULL, "UDP", 0, &__this_module, 0, 0, 
      &udp_print_private,
      &udp_check,
      &udp_bind,
      &udp_destroy }
};

static struct nf_setsockopt_ops ip_nat_udp_sockopts 
= { NULL, PF_INET, IP_NAT_UDP_ADD_USER, IP_NAT_UDP_DELETE_USER, do_udp_ctl };

int init_module(void)
{
	int ret;
	DEBUGP("ip_nat_udp\n");

	ret = ip_nat_register_proto(&proto_udp);
	if (ret < 0) return ret;

	ret = nf_register_sockopt(&ip_nat_udp_sockopts);
	if (ret < 0)
		ip_nat_unregister_proto(&proto_udp);

	return ret;
}

void cleanup_module()
{
	ip_nat_unregister_proto(&proto_udp);
	nf_unregister_sockopt(&ip_nat_udp_sockopts);
}
